Package: vecvec
Title: Construct Mixed Type Data Structures with Vectors of Vectors
Version: 1.1.0
Authors@R: c(
      person(
        given = "Mitchell",
        family = "O'Hara-Wild",
        role = c("aut", "cre"),
        email = "mail@mitchelloharawild.com",
        comment = c(ORCID = "0000-0001-6729-7695")
      ))
Description: Mixed type vectors are useful for combining semantically similar 
  classes. Some examples of semantically related classes include time across 
  different granularities (e.g. daily, monthly, annual) and probability 
  distributions (e.g. Normal, Uniform, Poisson). These groups of vector types 
  typically share common statistical operations which vary in results with the 
  attributes of each vector. The 'vecvec' data structure facilitates efficient 
  storage and computation across multiple vectors within the same object.
License: MIT + file LICENSE
Imports: rlang, S7, stats, vctrs
Suggests: lifecycle, testthat (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
URL: https://pkg.mitchelloharawild.com/vecvec/,
        https://github.com/mitchelloharawild/vecvec
BugReports: https://github.com/mitchelloharawild/vecvec/issues
Config/roxygen2/version: 7.3.3.9000
NeedsCompilation: no
Packaged: 2026-06-03 19:17:36 UTC; mitchell
Author: Mitchell O'Hara-Wild [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6729-7695>)
Maintainer: Mitchell O'Hara-Wild <mail@mitchelloharawild.com>
Repository: CRAN
Date/Publication: 2026-06-04 05:10:03 UTC
Built: R 4.6.0; ; 2026-06-04 09:26:51 UTC; unix

Package: vecvec
Title: Construct Mixed Type Data Structures with Vectors of Vectors
Version: 1.1.0
Authors@R: c(
      person(
        given = "Mitchell",
        family = "O'Hara-Wild",
        role = c("aut", "cre"),
        email = "mail@mitchelloharawild.com",
        comment = c(ORCID = "0000-0001-6729-7695")
      ))
Description: Mixed type vectors are useful for combining semantically similar 
  classes. Some examples of semantically related classes include time across 
  different granularities (e.g. daily, monthly, annual) and probability 
  distributions (e.g. Normal, Uniform, Poisson). These groups of vector types 
  typically share common statistical operations which vary in results with the 
  attributes of each vector. The 'vecvec' data structure facilitates efficient 
  storage and computation across multiple vectors within the same object.
License: MIT + file LICENSE
Imports: rlang, S7, stats, vctrs
Suggests: lifecycle, testthat (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
URL: https://pkg.mitchelloharawild.com/vecvec/,
        https://github.com/mitchelloharawild/vecvec
BugReports: https://github.com/mitchelloharawild/vecvec/issues
Config/roxygen2/version: 7.3.3.9000
NeedsCompilation: no
Packaged: 2026-06-03 19:17:36 UTC; mitchell
Author: Mitchell O'Hara-Wild [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6729-7695>)
Maintainer: Mitchell O'Hara-Wild <mail@mitchelloharawild.com>
Repository: CRAN
Date/Publication: 2026-06-04 05:10:03 UTC
Built: R 4.6.0; ; 2026-06-04 08:16:28 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(Complex,"vecvec::vecvec")
S3method(Math,"vecvec::vecvec")
S3method(Ops,"vecvec::vecvec")
S3method(Summary,"vecvec::vecvec")
S3method(`[<-`,"vecvec::vecvec")
S3method(`[[<-`,"vecvec::vecvec")
S3method(`dim<-`,"vecvec::vecvec")
S3method(`dimnames<-`,"vecvec::vecvec")
S3method(`is.na<-`,"vecvec::vecvec")
S3method(`length<-`,"vecvec::vecvec")
S3method(as.double,"vecvec::vecvec")
S3method(as.vector,"vecvec::vecvec")
S3method(vec_cast,"vecvec::vecvec")
S3method(vec_ptype2,"vecvec::vecvec.vecvec::vecvec")
export(class_vecvec)
export(cor)
export(cov)
export(is_vecvec)
export(unvecvec)
export(var)
export(vecvec)
export(vecvec_apply)
export(vecvec_apply_fn)
export(vecvec_register)
import(S7)
import(vctrs)
importFrom(rlang,expr)
importFrom(rlang,exprs)
importFrom(rlang,quo)
importFrom(rlang,quos)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(stats,median)

<!-- README.md is generated from README.qmd. Please edit that file -->

# vecvec <img src="man/figures/logo.svg" align="right" style="height: 139px;" alt="" />

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![R-CMD-check](https://github.com/mitchelloharawild/vecvec/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mitchelloharawild/vecvec/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/vecvec)](https://CRAN.R-project.org/package=vecvec)
<!-- badges: end -->

The vecvec package allows you to create vectors of mixed classes. Mixed
type vectors are useful for data analysis packages as they allow
semantically similar classes (such as time or distributions) to be
stored within the same vector and operated upon efficiently with the
same functions. It works similarly to `factor()`, in which the values of
a vector (analogous to levels of a factor) are stored in the attributes.
This also allows for efficient computation when the values are
replicated, since operations apply to the unique values of the vector.

This package is intended for developers creating mixed type vector
packages, and is unlikely useful for general data analysis (although I’d
be curious if you find direct application of this tool, let me know how
you use it).

## Installation

You can install the **stable** version from
[CRAN](https://cran.r-project.org/package=vecvec):

``` r
install.packages("vecvec")
```

You can install the development version of vecvec like so:

``` r
remotes::install_github("mitchelloharawild/vecvec")
```

## Example

Use `vecvec()` to create a vecvec. Here we combine three vectors of
different types.

``` r
library(vecvec)

x <- vecvec(as.Date("2025-03-29"), as.POSIXct("2025-01-03 10:05:00"), rnorm(10))
x
#> <vecvec[12]>
#>  [1] 2025-03-29          2025-01-03 10:05:00  1.262954285       
#>  [4] -0.326233361         1.329799263         1.272429321       
#>  [7]  0.414641434        -1.539950042        -0.928567035       
#> [10] -0.294720447        -0.005767173         2.404653389
```

Internally, the vectors are stored separately in the attributes.

``` r
str(x)
#> <vecvec::vecvec> int [1:12] 1 2 3 4 5 6 7 8 9 10 ...
#>  @ x:List of 3
#>  .. $ : Date[1:1], format: "2025-03-29"
#>  .. $ : POSIXct[1:1], format: "2025-01-03 10:05:00"
#>  .. $ : num [1:10] 1.263 -0.326 1.33 1.272 0.415 ...
```

Operations apply to all vector types.

``` r
x + 1
#> <vecvec[12]>
#>  [1] 2025-03-30          2025-01-03 10:05:01  2.26295428        
#>  [4]  0.67376664          2.32979926          2.27242932        
#>  [7]  1.41464143         -0.53995004          0.07143297        
#> [10]  0.70527955          0.99423283          3.40465339
```

When subsetting, unused values are dropped.

``` r
x[1]
#> <date*[1]>
#> [1] 2025-03-29
str(x[1])
#> <vecvec::vecvec> int 1
#>  @ x:List of 1
#>  .. $ : Date[1:1], format: "2025-03-29"
```

When a vecvec is replicated, only the indexing is replicated.

``` r
str(rep(x, each = 10))
#> <vecvec::vecvec> int [1:120] 1 1 1 1 1 1 1 1 1 1 ...
#>  @ x:List of 3
#>  .. $ : Date[1:1], format: "2025-03-29"
#>  .. $ : POSIXct[1:1], format: "2025-01-03 10:05:00"
#>  .. $ : num [1:10] 1.263 -0.326 1.33 1.272 0.415 ...
str(x[c(1,1,1,3,2,2)])
#> <vecvec::vecvec> int [1:6] 1 1 1 3 2 2
#>  @ x:List of 3
#>  .. $ : Date[1:1], format: "2025-03-29"
#>  .. $ : POSIXct[1:1], format: "2025-01-03 10:05:00"
#>  .. $ : num [1:10] 1.263 -0.326 1.33 1.272 0.415 ...
```

The values in the attribute are unchanged, which allows for efficient
computation since operations apply to the attributes (which are not
replicated).

``` r
str(rep(x, each = 10) + 100)
#> <vecvec::vecvec> int [1:120] 1 1 1 1 1 1 1 1 1 1 ...
#>  @ x:List of 3
#>  .. $ : Date[1:1], format: "2025-07-07"
#>  .. $ : POSIXct[1:1], format: "2025-01-03 10:06:40"
#>  .. $ : num [1:10] 101.3 99.7 101.3 101.3 100.4 ...
```

If the arguments is differ across replicated indices, the resulting
vecvec is
[*expanded*](https://github.com/mitchelloharawild/vecvec/issues/2)

``` r
str(rep(x, each = 10) + 1:120)
#> <vecvec::vecvec> int [1:120] 1 2 3 4 5 6 7 8 9 10 ...
#>  @ x:List of 3
#>  .. $ : Date[1:10], format: "2025-03-30" "2025-03-31" ...
#>  .. $ : POSIXct[1:10], format: "2025-01-03 10:05:11" "2025-01-03 10:05:12" ...
#>  .. $ : num [1:100] 22.3 23.3 24.3 25.3 26.3 ...
```

# vctrs compatibility methods

# proxy/restore
# This is inefficient but seemingly required for vctrs machinery
method(vec_proxy, class_vecvec) <- function(x, ...) {
  data_frame(x = list(x@x), l = sum(lengths(x@x)), i = S7_data(x))
}
method(vec_restore, class_vecvec) <- function(x, to, ...) {
  if (is_vecvec(x)) return(x)
  if (vec_size(x) == 0L) {
    return(S7_class(to)())
  }

  # Fast path for vecvec without merging - just update the indices
  if(identical(sum(slot_len <- lengths(to@x)), slot_grp <- unique(x$l))) {
    S7_data(to) <- x$i
    # Invoke pruning of slots
    out <- to[seq_along(to)]
    return(out)
  }
  
  # Identify groups of vectors
  # TODO - this destroys altrep
  grp <- data_frame(key = unique(x$x))
  grp$loc <- split(seq_along(x$i), match(x$x, grp$key))

  # val <- x$x[vapply(grp$loc, `[[`, integer(1), 1L)]
  # vctrs seems to drop list() to NULL internally somewhere? - fix it
  val <- lapply(grp$key, function(x) if (is.null(x)) list() else x)

  # Index offsets for each grp
  len <- c(
    0L,
    cumsum(vapply(val[-length(val)], function(x) sum(lengths(x)), integer(1)))
  )
  idx <- lapply(vec_seq_along(grp), function(i) x$i[grp$loc[[i]]] + len[i])

  # Restore the vecvec S7 object
  S7_class(to)(
    x = unlist(val, recursive = FALSE),
    i = unlist(idx, recursive = FALSE)
  )
}

# Comparison proxies
method(vec_proxy_equal, class_vecvec) <- function(x, ...) {
  # This is inefficient, but seems necessary for vctrs machinery.
  # Directly using `==` is faster as it applies on overlapping vctrs directly.
  data_frame(x = as.list(x), na = ifelse(is.na(x), NA, FALSE))
}
method(vec_proxy_compare, class_vecvec) <- function(x, ...) {
  xtfrm(x, ...)
}

# prototypes
vec_ptype2_vecvec <- function(x, y, ...) {
  if (!is_vecvec(x)) {
    x <- vecvec(x)
  }
  if (!is_vecvec(y)) {
    y <- vecvec(y)
  }

  x <- c(x, y)
  S7_data(x) <- integer()
  x
}
#' @export
`vec_ptype2.vecvec::vecvec.vecvec::vecvec` <- vec_ptype2_vecvec

# casting

#' Cast methods for vecvec types
#'
#' These functions are the underlying handlers for [vctrs::vec_cast()] when
#' casting to or from a `vecvec` type. They are not called directly; instead,
#' they are registered dynamically in [.onLoad()] as S3 methods covering every
#' `vec_cast.*` generic exported by **vctrs**.
#'
#' @param x The object to cast.
#' @param to The target ptype.
#' @param ... Passed on to inner [vctrs::vec_cast()] calls.
#'
#' @return
#' * `vec_cast_to_vecvec()` returns a `vecvec` object. Each element of `x` is
#'   cast to the ptype of the corresponding group in `to@x`.
#' * `vec_cast_from_vecvec()` returns an object of the type described by `to`,
#'   produced by [unvecvec()].
#'
#' @seealso [unvecvec()], [vctrs::vec_cast()]
#' @noRd
vec_cast_to_vecvec <- function(x, to, ...) {
  # If input and ptype have incompatible structure, produce flat vecvec type
  if (length(x) != length(to)) return(S7_class(to)(list(x)))
  
  # TODO - handle replicated indices
  if (anyDuplicated(S7_data(to))) {
    stop("Casting to vecvec with duplicated indices is not supported.", call. = FALSE)
  }

  # Match index positions and vec_cast the individual vectors
  idx <- S7_data(to)
  len <- c(0L, cumsum(lengths(to@x[-length(to@x)])))
  pos <- findInterval(idx, len, left.open = TRUE)
  loc <- vec_split(x, pos)
  to@x <- .mapply(
    function(i, val) vec_cast(val, to@x[[i]], ...),
    list(loc$key, loc$val), NULL
  )

  to
}
vec_cast_from_vecvec <- function(x, to, ...) {
  unvecvec(x, ptype = to)
}
#' @method vec_cast vecvec::vecvec
#' @export
`vec_cast.vecvec::vecvec` <- function(x, to, ...) {
  UseMethod("vec_cast.vecvec::vecvec", to)
}
# labels
method(vec_ptype_full, class_vecvec) <- function(x, ...) {
  if (length(x@x) != 1L) "vecvec" else paste0(vec_ptype_full(x@x[[1L]]), "*")
}
method(vec_ptype_abbr, class_vecvec) <- function(x, ...) {
  if (length(x@x) != 1L) "vecvec" else paste0(vec_ptype_abbr(x@x[[1L]]), "*")
}

.onLoad <- function(...) {
  S7::methods_register()

  # Register vecvec methods
  vecvec_register(class_vecvec)

  # Register all methods
  # lapply(
  #   setdiff(
  #     attr(methods(class = "default"), "info")$generic,
  #     attr(methods(class = "vecvec::vecvec"), "info")$generic
  #   ),
  #   register_s3_method,
  #   pkg = "vecvec", class = "vecvec::vecvec", 
  #   fun = vecvec_dispatch)
  invisible()
}

#' Apply a function to each vector of the vecvec
#'
#' The `vecvec_apply()` function applies a function `.f` to each vector in the 
#' `vecvec` vectors.
#' 
#' @param x A vecvec object
#' @param .f A function to apply to each vector
#' @param ... Additional arguments passed to `.f`
#' 
#' @return A vecvec data type with the same structure as `x` but with each 
#'   vector transformed by `.f`.
#' 
#' @export
vecvec_apply <- function(x, .f, ...) {
  if (vec_is_empty(x)) {
    return(x)
  }
  x@x <- vecvec_flatten_adj(lapply(x@x, .f, ...))

  x
}

#' Function factory for vecvec_apply
#' 
#' The `vecvec_apply_fn()` function is a function factory that creates applies
#' the function `.f` to each vector and optionally simplifies the result with 
#' `[unvecvec()]`. The function matches the forms of the original function 
#' `.f` and can be used to define methods for generic functions that apply to 
#' `vecvec` objects. If `.f` is a primitive function, the resulting function will
#' have a apply over an argument `x` and pass through `...`.
#' 
#' @inheritParams vecvec_apply
#' @param ptype A prototype to simplify to. If `NULL`, the result will be a
#'   `vecvec` object. If not `NULL`, the result will be simplified to the type 
#'   of `ptype` if possible.
#' @param SIMPLIFY If `TRUE`, the `[unvecvec()]` will be applied to the result 
#'   of `vecvec_apply()`, using `ptype` as the target type. If `FALSE`, the 
#'   result will always be a `vecvec` object.
#' 
#' @return A function that applies `.f` to each vector of a `vecvec` object and
#'  optionally simplifies the result.
#' 
#' @export
vecvec_apply_fn <- function(.f, ptype = NULL, SIMPLIFY = !is.null(ptype)) {
  fmls <- formals(args(.f))
  args <- names(fmls)[-1L]
  args <- `names<-`(syms(args), args)
  # Remove name of `...` if it exists
  arg_dots <- match("...", names(args), 0L)
  names(args)[arg_dots] <- ""

  apply_sym <- sym(names(fmls)[[1L]])
  apply_fn <- rlang::new_function(
    args = fmls,
    body = expr(vecvec::vecvec_apply(!!apply_sym, !!sym(".f"), !!!args)),
    env = rlang::new_environment(list(.f = .f), parent = rlang::caller_env())
  )

  if (SIMPLIFY) {
    body(apply_fn) <- expr(vecvec::unvecvec(!!body(apply_fn), ptype = !!ptype))
  }

  apply_fn
}

vecvec_dispatch <- function(x, ...) {
  x@x <- lapply(x@x, .Generic, ...)
  x
}

#' Compute overlap segments between two integer run-length vectors
#'
#' @description
#' Given two integer vectors `a` and `b` representing run lengths
#' along a common 1D axis (e.g. segments of contiguous positions),
#' this function computes the lengths of consecutive overlapping segments
#' between them. It assumes that `a` and `b` are positive integers and that
#' their sums are equal (i.e. lengths along same-sized vecvecs)
#'
#' Conceptually, you"“walk" along the sums of `a` and `b` simultaneously,
#' taking at each step the minimum of the remaining length in the current
#' element of `a` and `b`, and recording that as one overlap segment.
#'
#' @param a Integer vector of positive run lengths.
#' @param b Integer vector of positive run lengths.
#'
#' @return
#' An integer vector giving the lengths of consecutive overlapping segments.
#'
#' @details
#' This is an \eqn{O(length(a) + length(b))} algorithm using a single pass
#' through both vectors with constant additional memory, apart from the
#' preallocated output.
#'
#' @examples
#' a <- c(4L, 8L, 3L)
#' b <- c(3L, 9L, 3L)
#'
#' overlap_indices(a, b)
#' # [1] 3 1 8 3
#'
#' @noRd
overlap_indices <- function(a, b) {
  n_a <- length(a)
  n_b <- length(b)

  # Upper bound (very loose but safe)
  len <- i_a <- i_b <- integer(n_a + n_b - 1L)

  k <- 0L

  i <- 1L
  j <- 1L
  ra <- a[i]
  rb <- b[j]

  while (i <= n_a && j <= n_b) {
    x <- if (ra < rb) ra else rb
    k <- k + 1L
    len[k] <- x
    i_a[k] <- i
    i_b[k] <- j

    ra <- ra - x
    rb <- rb - x

    if (ra == 0L) {
      i <- i + 1L
      if (i <= n_a) ra <- a[i]
    }
    if (rb == 0L) {
      j <- j + 1L
      if (j <= n_b) rb <- b[j]
    }
  }

  # Trim to actual used length
  list(
    len = len[seq_len(k)],
    idx = list(
      i_a = i_a[seq_len(k)],
      i_b = i_b[seq_len(k)]
    )
  )
}

# This causes fragmentation as new values are appended to vector list
# A defragmentation function would be useful (#18)
# @method [<- vecvec::vecvec
#' @rawNamespace S3method(`[<-`,"vecvec::vecvec")
`[<-.vecvec::vecvec` <- function(x, i, ..., value) {
  # Recycle `value` to the length of `i`
  replacements <- S7_data(x)[i, ...]
  value <- vec_recycle(value, size = length(replacements))

  # Remove unreferenced values from `x@x`
  vec_rm <- unique(replacements)
  if (is.array(x)) {
    # Also remove out-of-bounds indices produced
    # by array() when length(x) != prod(dim(x))
    nx <- sum(lengths(x@x))
    if (length(x) < nx) {
      vec_rm <- c(vec_rm, seq(from = length(x) + 1L, to = nx))
    }
  }
  vec_starts <- c(0L, cumsum(lengths(x@x)[-length(x@x)]))
  vec_idx <- findInterval(vec_rm, vec_starts, left.open = TRUE)
  vec_pos <- vec_split(vec_rm - vec_starts[vec_idx], vec_idx)
  for (k in seq_len(nrow(vec_pos))) {
    x@x[[vec_pos$key[[k]]]] <- x@x[[vec_pos$key[[k]]]][-vec_pos$val[[k]]]
  }

  # Append `value` to the end of `x@x` and update `x@i` to point to the new values
  if (is_vecvec(value)) {
    x@x <- c(x@x, value@x)
    idx <- S7_data(value) + length(x)
  } else {
    x@x <- c(x@x, list(value))
    idx <- seq_along(value) + length(x)
  }
  S7_data(x)[i, ...] <- idx
  
  # Rank local indices
  S7_data(x) <- vec_rank(as.integer(S7_data(x)))

  # Crude but fast defragmentation of adjacent same-type vectors
  vecvec_flatten_adj(x)
}

#' @method [[<- vecvec::vecvec
#' @rawNamespace S3method(`[[<-`,"vecvec::vecvec")
`[[<-.vecvec::vecvec` <- function(x, i, value) {
  if (length(i) != 1L) {
    stop("attempt to replace multiple elements in a vecvec", call. = FALSE)
  }

  x[i] <- value
  x
}

# The vector values are set to NA rather than the vecvec index being set to NA.

# @method is.na<- vecvec::vecvec
#' @rawNamespace S3method(`is.na<-`,"vecvec::vecvec")
`is.na<-.vecvec::vecvec` <- function(x, value) {
  if (is.logical(value)) {
    if (length(value) != length(x)) {
      stop("logical subscript has wrong length", call. = FALSE)
    }
    value <- which(value)
  }

  # Replace values with NA
  vec_na <- unique(S7_data(x)[value])
  vec_starts <- c(0L, cumsum(lengths(x@x)[-length(x@x)]))
  vec_idx <- findInterval(vec_na, vec_starts, left.open = TRUE)
  vec_pos <- vec_split(vec_na - vec_starts[vec_idx], vec_idx)
  for (k in seq_len(nrow(vec_pos))) {
    x@x[[vec_pos$key[[k]]]][vec_pos$val[[k]]] <- NA
  }

  x
}

method(mean, class_vecvec) <- function(x, ...) {
  mean(unvecvec(x), ...)
}

method(
  weighted.mean,
  class_vecvec
) <- function(x, w, ...) {
  weighted.mean(unvecvec(x), w = w, ...)
}

#' @importFrom stats median
method(
  median,
  class_vecvec
) <- function(x, ...) {
  median(unvecvec(x), ...)
}

method(
  quantile, 
  class_vecvec
) <- function(x, ...) {
  quantile(unvecvec(x), ...)
}

# Mask the [stats::cor] family
#' Correlation, Variance and Covariance
#' 
#' @description
#' 
#' `var`, `cov`, and `cor`  compute the variance of x and the covariance or 
#' correlation of x and y if these are vectors. If x and y are matrices then 
#' the covariances (or correlations) between the columns of x and the columns 
#' of y are computed.
#' 
#' These functions mask and wrap the [`stats::var`], [`stats::cov`], and 
#' [`stats::cor`] functions to add support for `vecvec` objects. More details
#' can be found in the documentation for those functions.
#' 
#' @inheritParams stats::cor
#' @param ... Additional arguments passed to [`stats::var`], [`stats::cov`], and 
#' [`stats::cor`].
#' 
#' @seealso [stats::var()], [stats::cov()], [stats::cor()]
#' 
#' @name masked-cor
#' @export
var <- new_generic("var", "x")
method(var, class_vecvec) <- function(x, y = NULL, ...) {
  if (is_vecvec(y)) y <- unvecvec(y)
  var(unvecvec(x), y = y, ...)
}
method(var, class_any) <- function(x, ...) {
  stats::var(x, ...)
}

#' @name masked-cor
#' @export
cov <- new_generic("cov", "x")
method(cov, class_vecvec) <- function(x, y = NULL, ...) {
  if (is_vecvec(y)) y <- unvecvec(y)
  cov(unvecvec(x), y = y, ...)
}
method(cov, class_any) <- function(x, ...) {
  stats::cov(x, ...)
}

#' @name masked-cor
#' @export
cor <- new_generic("cor", "x")
method(cor, class_vecvec) <- function(x, y = NULL, ...) {
  if (is_vecvec(y)) y <- unvecvec(y)
  cor(unvecvec(x), y = y, ...)
}
method(cor, class_any) <- function(x, ...) {
  stats::cor(x, ...)
}

# S7-style implementation pending https://github.com/RConsortium/S7/issues/353
# Indicative release in S7 v0.3.0
#' @method Ops vecvec::vecvec
#' @export
`Ops.vecvec::vecvec` <- function(e1, e2) {
  bool_op <- switch(
    .Generic,
    `<` = ,
    `>` = ,
    `==` = ,
    `!=` = ,
    `<=` = ,
    `>=` = TRUE,
    FALSE
  )

  # Unary operation (shortcut method on attributes)
  if (missing(e2)) {
    e1@x <- lapply(e1@x, .Generic)
    return(e1)
  }

  # Binary operation (complete method on values)
  # Convert to vecvec before recycling so vctrs repeats indices rather than
  # expanding the underlying vector (which would defeat compression).
  if (!is_vecvec(e1)) e1 <- vecvec(e1) else class_vv <- S7_class(e1)
  if (!is_vecvec(e2)) e2 <- vecvec(e2) else class_vv <- S7_class(e2)
  args <- vec_recycle_common(e1 = e1, e2 = e2)
  n <- vec_size(args[[1L]])

  # Cumulative slot boundaries: map indices -> (slot index, within-slot position).
  b <- lapply(args, function(a) c(0L, cumsum(lengths(a@x))))
  slot <- lapply(
    seq_along(args),
    function(i) findInterval(S7_data(args[[i]]), b[[i]], left.open = TRUE)
  )
  within <- .mapply(
    function(arg, b, slot) S7_data(arg) - b[slot],
    list(args, b, slot), NULL
  )

  # Encode the (slot$e1, slot$e2) pair as a single key; contiguous runs define output slots.
  slot_pair_key <- (slot[[1L]] - 1L) * (length(args$e2@x) + 1L) + slot[[2L]]
  out_slot <- cumsum(c(TRUE, slot_pair_key[-1L] != slot_pair_key[-n]))
  n_slots <- out_slot[length(out_slot)]

  # For each output slot: compute the op on unique (w$e1, w$e2) pairs and record val_idx.
  # Returns list(vals, val_idx) so result_i can be built without a second pass.
  groups <- split(seq_len(n), out_slot)
  computed <- lapply(groups, function(pos) {
    w <- lapply(within, `[`, pos)
    s_idx <- lapply(slot, `[[`, pos[1L])

    pair_key <- (w[[1L]] - 1L) * length(args[[2L]]@x[[s_idx[[2L]]]]) + w[[2L]]
    unique_keys <- unique(pair_key)
    first_of <- match(unique_keys, pair_key)

    list(
      vals = do.call(
        .Generic,
        .mapply(
          function(arg, s, w) arg@x[[s]][w][first_of],
          list(args, s_idx, w), NULL
        )
      ),
      val_idx = match(pair_key, unique_keys)
    )
  })

  result_x <- unname(lapply(computed, `[[`, "vals"))

  # Precompute per-slot offsets once, then scatter val_idx into result_i.
  offsets <- c(0L, cumsum(lengths(result_x))[-n_slots])
  result_i <- integer(n)
  for (s in seq_len(n_slots)) {
    result_i[groups[[s]]] <- offsets[[s]] + computed[[s]]$val_idx
  }

  # TODO - use a method to identify if a better class can be returned.
  res <- class_vv(result_x, result_i)

  if (bool_op) {
    res <- unvecvec(res)
  }
  res
}

#' @method Math vecvec::vecvec
#' @export
`Math.vecvec::vecvec` <- function(x, ...) {
  if (.Generic %in% c("cumsum, cumprod, cummax, cummin")) {
    rlang::abort("Culumative operations are not yet supported")
  }
  x@x <- lapply(x@x, .Generic, ...)
  # TODO - Detect if all listed prototypes are compatible, then collapse if flat
  x
}

#' @method Summary vecvec::vecvec
#' @export
`Summary.vecvec::vecvec` <- function(..., na.rm = FALSE) {
  do.call(
    .Generic,
    c(
      lapply(
        rlang::list2(...),
        function(x) {if (is_vecvec(x)) unvecvec(x) else x}
      ),
      list(na.rm = na.rm)
    )
  )
}

#' @method Complex vecvec::vecvec
#' @export
`Complex.vecvec::vecvec` <- function(z) {
  z@x <- lapply(z@x, .Generic)
  unvecvec(z)
}

method(unique, class_vecvec) <- function(x, incomparables = FALSE, ...) {
  x[!duplicated(x)]
}

method(diff, class_vecvec) <- function (x, lag = 1L, differences = 1L, ...) 
{
  ismat <- is.matrix(x)
  if (length(lag) != 1L || length(differences) != 1L || lag < 1L || differences < 1L)
  	stop("'lag' and 'differences' must be integers >= 1")
  i1 <- -seq_len(lag)
  if (ismat) for (i in seq_len(differences))
    x <- x[i1, , drop = FALSE] - x[seq_len(max(nrow(x) - lag, 0L)), , drop = FALSE] 
  else for (i in seq_len(differences))
    x <- x[i1] - `length<-`(x, max(length(x) - lag, 0L))
  x
}

vecvec_flatten_adj <- function(x) {
  if (is_vecvec(x)) {
    x@x <- vecvec_flatten_adj(x@x)
    return(x)
  }

  # Vector prototypes
  ptypes <- lapply(x, `[`, 0L)

  i <- 1L
  while (i < length(x)) {
    if (identical(ptypes[[i]], ptypes[[i + 1L]])) {
      x[[i]] <- c(x[[i]], x[[i + 1L]])
      x <- x[-(i + 1L)]
      ptypes <- ptypes[-(i + 1L)]
    } else {
      i <- i + 1L
    }
  }

  x
}

# Internal helper for casting which is less strict than `vec_cast()` for 
# user specified casts. This is used in `as.*` methods and `unvecvec()` when a
# `ptype` is supplied.
vec_cast_unsafe <- S7::new_generic("vec_cast_unsafe", c("x", "to"))
method(vec_cast_unsafe, list(class_any, class_vecvec)) <- function(x, to) {
  vecvec(x)
}
method(vec_cast_unsafe, list(class_any, class_logical)) <- function(x, to) as.logical(x)
method(vec_cast_unsafe, list(class_any, class_integer)) <- function(x, to) as.integer(x)
method(vec_cast_unsafe, list(class_any, class_double)) <- function(x, to) as.double(x)
method(vec_cast_unsafe, list(class_any, class_complex)) <- function(x, to) as.complex(x)
method(vec_cast_unsafe, list(class_any, class_character)) <- function(x, to) as.character(x)
method(vec_cast_unsafe, list(class_any, class_raw)) <- function(x, to) as.raw(x)
method(vec_cast_unsafe, list(class_any, class_Date)) <- function(x, to) as.Date(x)
method(vec_cast_unsafe, list(class_any, class_POSIXct)) <- function(x, to) as.POSIXct(x, tz = attr(to, "tzone", exact = TRUE))
method(vec_cast_unsafe, list(class_any, class_POSIXlt)) <- function(x, to) as.POSIXlt(x, tz = attr(to, "tzone", exact = TRUE))
method(vec_cast_unsafe, list(class_any, class_any)) <- function(x, to) vec_cast(x, to)

globalVariables("properties")

# Casting methods
# method(convert, list(class_any, class_vecvec)) <- function(from, to, ...) {
#   vecvec(from)
# }
# method(
#   convert,
#   list(
#     class_vecvec,
#     new_union(
#       class_vector,
#       class_Date,
#       class_POSIXct,
#       class_POSIXlt,
#       class_data.frame
#     )
#   )
# ) <- function(from, to, ...) unvecvec(from, ptype = to$constructor())

method(as.logical, class_vecvec) <- function(x, ...) unvecvec(x, ptype = class_logical$constructor())
method(as.raw, class_vecvec) <- function(x, ...) unvecvec(x, ptype = class_raw$constructor())
method(as.integer, class_vecvec) <- function(x, ...) unvecvec(x, ptype = class_integer$constructor())
# For some reason this method isn't being registered properly with S7::method<-
# method(as.double, class_vecvec) <- function(x) unvecvec(x, ptype = class_double$constructor())
#' @export
#' @method as.double vecvec::vecvec
`as.double.vecvec::vecvec` <- function(x, ...) unvecvec(x, ptype = class_double$constructor())
method(as.complex, class_vecvec) <- function(x, ...) unvecvec(x, ptype = class_complex$constructor())
method(as.character, class_vecvec) <- function(x, ...) unvecvec(x, ptype = class_character$constructor())
method(as.Date, class_vecvec) <- function(x, ...) unvecvec(x, ptype = class_Date$constructor())
method(as.data.frame, class_vecvec) <- function(x, ...) unvecvec(x, ptype = class_data.frame$constructor())
method(as.POSIXct, class_vecvec) <- function(x, tz = "", ...) {
  unvecvec(x, ptype = class_POSIXct$constructor(tz = tz))
}
method(as.POSIXlt, class_vecvec) <- function(x, tz = "", ...) {
  unvecvec(x, ptype = class_POSIXlt$constructor(tz = tz))
}
method(as.list, class_vecvec) <- function(x, ...) {
  x@x <- lapply(x@x, as.list)
  unvecvec(x, ptype = list())
}

#' Create a vector of vectors
#'
#' @description
#' 
#' `r lifecycle::badge('stable')`
#'
#' A `vecvec` is a vector that holds elements of different types without
#' coercing them to a common type. Unlike a list of vectors, a `vecvec` behaves 
#' as a flat vector (hence vector of vectors). This means that you can 
#' operations (such as indexing, arithmetic, and statistics) apply across all
#' elements of a `vecvec` as if they were combined into a single vector.
#'
#' Mixed-type vectors can be useful if you need to store heterogeneous data in a
#' vector-like structure. In most cases, I believe this is bad practice for data
#' analytics, but this could be useful for tidying up messy data. The most
#' valuable use case for vecvec is as a data structure for mixed-type semantic
#' vectors. This package is used by [mixtime](https://pkg.mitchelloharawild.com/mixtime/)
#' and [distributional](https://pkg.mitchelloharawild.com/distributional/) to 
#' create vectors of time with different chronons and distributions with 
#' different shapes.
#'
#' To convert a `vecvec` back to a plain typed vector, use [unvecvec()], which
#' casts all elements to a common type via [vctrs::vec_cast()].
#'
#' @param ... Vectors to combine. Each vector is stored as a separate typed
#'   slot; no type coercion is performed.
#'
#' @return A `vecvec` object whose length equals the total number of elements
#'   across all input vectors.
#'
#' @seealso [unvecvec()] to coerce a `vecvec` to a single-typed vector;
#'   [is_vecvec()] to test whether an object is a `vecvec`.
#'
#' @examples
#' # Mixed types are preserved without coercion
#' vv <- vecvec(Sys.Date(), rnorm(3), letters)
#' vv
#'
#' # .[i] Indexing works like a flat vector
#' vv[c(1L, 3L, 7L)]
#' 
#' # .[[i]] drops to the original vector type
#' vv[[2L]]
#'
#' @export
vecvec <- function(...) {
  class_vecvec(x = rlang::list2(...))
}

#' Coerce a vector of vectors to a single typed vector
#'
#' @description
#' 
#' `r lifecycle::badge('stable')`
#'
#' Flattens a `vecvec` into a plain R vector by casting all elements to a
#' common type. This is the inverse of [vecvec()].
#'
#' Type resolution follows vctrs coercion rules: when `ptype` is `NULL` the
#' common type is determined automatically from the slots of `x` via
#' [vctrs::vec_ptype_common()]. If no common type can be found (e.g. all slots
#' are empty), the result falls back to `logical()`.
#'
#' @param x A `vecvec` object.
#' @param ptype A prototype specifying the desired output type, e.g.
#'   `character()` or `numeric()`. If `NULL` (the default), the common type is
#'   inferred from the elements of `x` using [vctrs::vec_ptype_common()],
#'   falling back to `logical()` when no common type can be determined. Passing
#'   an explicit `ptype` is useful when you need a guaranteed output type
#'   regardless of what `x` contains, or when automatic inference would pick an
#'   undesirable type.
#'
#' @return A vector of length `length(x)` and type `ptype` (or the inferred
#'   common type when `ptype = NULL`). Positions corresponding to `NA` indices
#'   in the underlying `vecvec` structure are filled with `NA`.
#'
#' @seealso [vecvec()] to create a `vecvec`; [vctrs::vec_ptype_common()] for
#'   the type inference rules; [vctrs::vec_cast()] for the casting rules.
#'
#' @examples
#' vv <- vecvec(1:3, c(4.5, 5.5))
#'
#' # Automatic type inference: integer + double -> double
#' unvecvec(vv)
#'
#' # Force a specific output type
#' unvecvec(vv, ptype = character())
#'
#' @export
unvecvec <- function(x, ptype = NULL) {
  if (!is_vecvec(x)) {
    stop("`x` must be a vecvec object", call. = FALSE)
  }

  if ((len <- length(x)) == 0L) {
    return(unlist(x@x))
  }

  # Cast mixed vector types to common type
  if (is.null(ptype)) {
    ptype <- vec_ptype_common(!!!x@x) %||% logical()
    # Use safe casting from vctrs
    vec_cast_unsafe <- vec_cast
  }
  x@x <- lapply(x@x, vec_cast_unsafe, to = ptype)

  # Construct output single-typed vector
  res <- vec_init(ptype, n = len)
  pos <- !is.na(S7_data(x))
  res[pos] <- vec_c(!!!x@x)[S7_data(x)[pos]]
  res
}

# ------------------------------------------------------------------------------
# Display methods
# ------------------------------------------------------------------------------
method(print, class_vecvec) <- function(x, ...) {
  vctrs::obj_print(x, ...)
}
method(format, class_vecvec) <- function(x, ...) {
  fmt <- vec_c(!!!lapply(x@x, format, ...), .ptype = character())[S7_data(x)]
  dim(fmt) <- dim(x)
  dimnames(fmt) <- dimnames(x)
  fmt
}

# ------------------------------------------------------------------------------
# Attribute methods
# ------------------------------------------------------------------------------
# @method length<- vecvec::vecvec
#' @rawNamespace S3method(`length<-`,"vecvec::vecvec")
`length<-.vecvec::vecvec` <- function(x, value) x[seq_len(value)]

#' Test if an object is a vecvec
#'
#' @param x Object to test.
#'
#' @return `TRUE` if `x` inherits from `class_vecvec`, `FALSE` otherwise.
#'
#' @seealso [vecvec()] to create a vecvec object.
#'
#' @examples
#' vv <- vecvec(1:3, letters)
#' is_vecvec(vv)
#' is_vecvec(1:3)
#'
#' @export
is_vecvec <- function(x) S7_inherits(x, class_vecvec)

# ------------------------------------------------------------------------------
# Indexing methods
# ------------------------------------------------------------------------------
method(`[`, class_vecvec) <- function(x, i, ...) {
  idx <- S7_data(x)[i, ...]
  not_na <- !is.na(idx)

  if (!any(not_na)) {
    return(S7_class(x)(x = list(), i = idx))
  }

  idx_nn <- idx[not_na]

  # Slot start positions in original x@x
  orig_starts <- c(0L, cumsum(lengths(x@x[-length(x@x)])))

  # Which original slot each selected element belongs to
  pos <- findInterval(idx_nn, orig_starts, left.open = TRUE)

  # Drop entirely unreferenced slots
  keep <- sort(unique(pos))
  all_kept <- length(keep) == length(x@x)

  x@x <- x@x[keep]
  orig_starts <- orig_starts[keep]

  new_slot <- match(pos, keep)
  local_idx <- idx_nn - orig_starts[new_slot] # 1-based within slot

  # TODO - remove unreferenced values from x@x if *all* references are dropped
  if (!anyDuplicated(idx_nn)) {
    slot_lengths <- lengths(x@x)
    groups <- split(seq_along(new_slot), new_slot)

    for (k in seq_along(x@x)) {
      el <- groups[[k]]
      sel <- local_idx[el]
      u <- unique(sel)
      if (length(u) < slot_lengths[k]) {
        local_idx[el] <- match(sel, u)
        x@x[[k]] <- x@x[[k]][u]
      }
    }
  }

  # Assign new vecvec indices
  new_starts <- c(0L, cumsum(lengths(x@x[-length(x@x)])))
  idx[not_na] <- new_starts[new_slot] + local_idx
  dim(x) <- NULL
  S7_data(x) <- idx
  dim(x) <- dim(idx)

  x
}

method(`[[`, class_vecvec) <- function(x, i, ...) {
  idx <- S7_data(x)[i]
  len <- c(0L, cumsum(lengths(x@x[-length(x@x)])))
  pos <- findInterval(idx, len, left.open = TRUE)
  x@x[[pos]][[idx - len[pos]]]
}

# ------------------------------------------------------------------------------
# Combining methods
# ------------------------------------------------------------------------------
method(c, class_vecvec) <- function(..., recursive = FALSE) {
  dots <- unname(rlang::list2(...))
  is_vv <- vapply(dots, is_vecvec, logical(1L))
  dots[!is_vv] <- lapply(dots[!is_vv], vecvec)

  # TODO - reduce structure into a common vecvec
  i_offsets <- cumsum(vapply(
    dots[-length(dots)],
    function(x) sum(lengths(x@x)),
    integer(1L)
  ))

  x <- lapply(dots, function(x) x@x)
  i <- c(
    S7_data(dots[[1L]]),
    .mapply(
      function(x, j) S7_data(x) + j,
      list(x = dots[-1L], j = i_offsets),
      NULL
    )
  )

  S7_class(..1)(
    x = unlist(x, recursive = FALSE),
    i = unlist(i, recursive = FALSE)
  )
}

# ------------------------------------------------------------------------------
# Array methods
# ------------------------------------------------------------------------------

# @method as.vector vecvec::vecvec
#' @rawNamespace S3method(as.vector,"vecvec::vecvec")
`as.vector.vecvec::vecvec` <- function(x, mode = "any") x

# @method dim<- vecvec::vecvec
#' @rawNamespace S3method(`dim<-`,"vecvec::vecvec")
`dim<-.vecvec::vecvec` <- function(x, value) {
  attr(x, "dim") <- value
  x
}
# @method dimnames<- vecvec::vecvec
#' @rawNamespace S3method(`dimnames<-`,"vecvec::vecvec")
`dimnames<-.vecvec::vecvec` <- function(x, value) {
  attr(x, "dimnames") <- value
  x
}

test_that("sort() increasing order", {
  rand <- rnorm(5, sd = 5)
  expect_equal(
    as.numeric(sort(vecvec(1:10, rand))),
    sort(c(1:10, rand))
  )
  expect_equal(
    as.numeric(sort(sample(vecvec(1:10, rand)))),
    sort(c(1:10, rand))
  )
})

test_that("sort() decreasing order", {
  vv_num <- vecvec(c(1.5, 3.5, 2.5, 5.5, 4.5), c(1, 3, 2, 5, 4))
  expect_equal(
    as.numeric(sort(vv_num, decreasing = TRUE)),
    sort(as.numeric(vv_num), decreasing = TRUE)
  )
})

test_that("Unary Ops", {
  expect_equal(
    as.numeric(-vecvec(1:10)),
    -1:-10
  )
  expect_equal(
    as.numeric(+vecvec(1:10)),
    1:10
  )
})

test_that("Binary Ops with single segments", {
  expect_equal(
    as.numeric(vecvec(1:10) + 1:10),
    seq(2,20, by = 2)
  )
  expect_equal(
    as.numeric(vecvec(1:10) + vecvec(1:10)),
    seq(2,20, by = 2)
  )
})

test_that("Binary Ops with scalars", {
  expect_equal(
    as.numeric(vecvec(1:10) + 5),
    6:15
  )

  rand <- rbeta(5, 3, 7)
  expect_equal(
    vecvec(-5:9, rand) >= 0,
    c(rep(FALSE, 5), rep(TRUE, 15))
  )
})

test_that("Binary Ops with overlapping segments", {
  vecvec1 <- vecvec(1:10, 11:20)
  vecvec2 <- vecvec(5:14, 15:24)
  expect_equal(
    as.numeric(vecvec1 + vecvec2),
    1:20 + 5:24
  )
})

test_that("Binary Ops with non-overlapping segments", {
  vecvec1 <- vecvec(1:5, 6:20)
  vecvec2 <- vecvec(21:30, 31:40)
  expect_equal(
    as.numeric(vecvec1 + vecvec2),
    1:20 + 21:40
  )
})

test_that("Binary Ops don't expand when arguments are invariable over replicates", {
  x <- vecvec(1:10, as.double(11:20), 21:30)

  result <- rep(x, each = 10) + 100L
  result_vals <- as.numeric(result)

  expected <- rep(c(1:10, 11:20, 21:30), each = 10) + 100
  expect_equal(result_vals, expected)

  expect_equal(
    lengths(result@x),
    c(10L, 10L, 10L)
  )
})

test_that("Binary Ops expand when arguments differ across replicated indices", {
  x <- vecvec(1:10, as.double(11:20), 21:30)

  result <- rep(x, each = 10) + 1:300
  result_vals <- as.numeric(result)

  expected <- rep(c(1:10, 11:20, 21:30), each = 10) + 1:300
  expect_equal(result_vals, expected)
  expect_equal(
    lengths(result@x),
    c(100L, 100L, 100)
  )
})

# Statistics: mean(), weighted.mean(), median(), var(), quantile(), cov(), cor()

# --- mean() -----------------------------------------------------------------

test_that("mean() of a single-segment double vecvec matches base mean()", {
  x <- c(1.5, 2.5, 3.5, 4.5)
  expect_equal(mean(vecvec(x)), mean(x))
})

test_that("mean() of a multi-segment vecvec matches base mean()", {
  x <- 1:5
  y <- 6:10
  expect_equal(mean(vecvec(x, y)), mean(c(x, y)))
})

test_that("mean() of an integer vecvec returns a double", {
  expect_equal(mean(vecvec(1L, 2L, 3L)), 2)
})

test_that("mean() with NA returns NA by default", {
  x <- c(1, NA, 3)
  expect_identical(mean(vecvec(x)), NA_real_)
})

test_that("mean() respects na.rm = TRUE", {
  x <- c(1, NA, 3)
  expect_equal(mean(vecvec(x), na.rm = TRUE), mean(x, na.rm = TRUE))
})

# --- weighted.mean() --------------------------------------------------------

test_that("weighted.mean() matches base weighted.mean() for equal weights", {
  x <- c(1, 2, 3, 4, 5)
  w <- c(1, 1, 1, 1, 1)
  expect_equal(weighted.mean(vecvec(x), w = w), weighted.mean(x, w = w))
})

test_that("weighted.mean() matches base weighted.mean() for unequal weights", {
  x <- c(1, 2, 3, 4, 5)
  w <- c(0.1, 0.2, 0.3, 0.2, 0.2)
  expect_equal(weighted.mean(vecvec(x), w = w), weighted.mean(x, w = w))
})

test_that("weighted.mean() of a multi-segment vecvec matches base", {
  x <- c(1.0, 2.0, 3.0)
  y <- c(4.0, 5.0)
  w <- c(0.1, 0.2, 0.3, 0.2, 0.2)
  expect_equal(
    weighted.mean(vecvec(x, y), w = w),
    weighted.mean(c(x, y), w = w)
  )
})

test_that("weighted.mean() respects na.rm = TRUE", {
  x <- c(1, NA, 3)
  w <- c(1, 1, 1)
  expect_equal(
    weighted.mean(vecvec(x), w = w, na.rm = TRUE),
    weighted.mean(x, w = w, na.rm = TRUE)
  )
})

# --- median() ---------------------------------------------------------------

test_that("median() of an odd-length integer vecvec matches base median()", {
  x <- c(3L, 1L, 4L, 1L, 5L)
  expect_equal(median(vecvec(x)), median(x))
})

test_that("median() of an even-length double vecvec matches base median()", {
  x <- c(1.0, 2.0, 3.0, 4.0)
  expect_equal(median(vecvec(x)), median(x))
})

test_that("median() of a multi-segment vecvec matches base median()", {
  x <- 1:5
  y <- 6:10
  expect_equal(median(vecvec(x, y)), median(c(x, y)))
})

test_that("median() with NA returns NA by default", {
  x <- c(1, NA, 3)
  expect_identical(median(vecvec(x)), NA_real_)
})

test_that("median() respects na.rm = TRUE", {
  x <- c(1, NA, 3, 5)
  expect_equal(median(vecvec(x), na.rm = TRUE), median(x, na.rm = TRUE))
})

# --- var() ------------------------------------------------------------------

test_that("var() of a single-segment double vecvec matches base var()", {
  x <- c(2.0, 4.0, 4.0, 4.0, 5.0, 5.0, 7.0, 9.0)
  expect_equal(var(vecvec(x)), var(x))
})

test_that("var() of an integer vecvec matches base var()", {
  x <- 1:10
  expect_equal(var(vecvec(x)), var(x))
})

test_that("var() of a multi-segment vecvec matches base var()", {
  x <- c(1.0, 2.0, 3.0)
  y <- c(4.0, 5.0, 6.0)
  expect_equal(var(vecvec(x, y)), var(c(x, y)))
})

test_that("var() of a length-1 vecvec returns NA", {
  expect_identical(var(vecvec(42)), NA_real_)
})

test_that("var() with NA returns NA by default", {
  x <- c(1, NA, 3)
  expect_identical(var(vecvec(x)), NA_real_)
})

test_that("var() respects na.rm = TRUE", {
  x <- c(1, NA, 3, 4)
  expect_equal(var(vecvec(x), na.rm = TRUE), var(x, na.rm = TRUE))
})

# --- quantile() -------------------------------------------------------------

test_that("quantile() default probs match base quantile()", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  expect_equal(quantile(vecvec(x)), quantile(x))
})

test_that("quantile() with custom probs matches base quantile()", {
  x <- 1:100
  probs <- c(0.1, 0.25, 0.5, 0.75, 0.9)
  expect_equal(quantile(vecvec(x), probs = probs), quantile(x, probs = probs))
})

test_that("quantile() of a multi-segment vecvec matches base quantile()", {
  x <- c(1.0, 3.0, 5.0)
  y <- c(2.0, 4.0, 6.0)
  expect_equal(quantile(vecvec(x, y)), quantile(c(x, y)))
})

test_that("quantile() respects names = FALSE", {
  x <- 1:10
  expect_equal(
    quantile(vecvec(x), names = FALSE),
    quantile(x, names = FALSE)
  )
})

test_that("quantile() respects na.rm = TRUE", {
  x <- c(1, NA, 3, 4, 5)
  expect_equal(
    quantile(vecvec(x), na.rm = TRUE),
    quantile(x, na.rm = TRUE)
  )
})

test_that("quantile() with a single prob returns a named scalar", {
  x <- 1:10
  result <- quantile(vecvec(x), probs = 0.5)
  expect_equal(result, quantile(x, probs = 0.5))
})

# --- cov() ------------------------------------------------------------------

test_that("cov() with vecvec x and plain vector y matches base cov()", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  y <- c(2.0, 4.0, 5.0, 4.0, 5.0)
  expect_equal(cov(vecvec(x), y = y), cov(x, y = y))
})

test_that("cov() with both x and y as vecvec matches base cov()", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  y <- c(5.0, 4.0, 3.0, 2.0, 1.0)
  expect_equal(cov(vecvec(x), y = vecvec(y)), cov(x, y = y))
})

test_that("cov() of a multi-segment vecvec x matches base cov()", {
  x <- c(1.0, 2.0, 3.0)
  y_seg <- c(4.0, 5.0, 6.0)
  y <- c(2.0, 4.0, 6.0, 5.0, 3.0, 1.0)
  expect_equal(cov(vecvec(x, y_seg), y = y), cov(c(x, y_seg), y = y))
})

test_that("cov() respects use = 'complete.obs' with NAs", {
  x <- c(1, NA, 3, 4, 5)
  y <- c(2, 4,  5, 4, 5)
  expect_equal(
    cov(vecvec(x), y = y, use = "complete.obs"),
    cov(x, y = y, use = "complete.obs")
  )
})

# --- cor() ------------------------------------------------------------------
test_that("cor() with vecvec x and plain vector y matches base cor()", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  y <- c(2.0, 4.0, 5.0, 4.0, 5.0)
  expect_equal(cor(vecvec(x), y = y), cor(x, y = y))
})

test_that("cor() with both x and y as vecvec matches base cor()", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  y <- c(5.0, 4.0, 3.0, 2.0, 1.0)
  expect_equal(cor(vecvec(x), y = vecvec(y)), cor(x, y = y))
})

test_that("cor() of perfectly correlated vectors returns 1", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  expect_equal(cor(vecvec(x), y = x), 1)
})

test_that("cor() of perfectly anti-correlated vectors returns -1", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  y <- rev(x)
  expect_equal(cor(vecvec(x), y = y), -1)
})

test_that("cor() of a multi-segment vecvec x matches base cor()", {
  x <- c(1.0, 2.0, 3.0)
  y_seg <- c(4.0, 5.0, 6.0)
  y <- c(2.0, 4.0, 6.0, 5.0, 3.0, 1.0)
  expect_equal(cor(vecvec(x, y_seg), y = y), cor(c(x, y_seg), y = y))
})

test_that("cor() respects use = 'complete.obs' with NAs", {
  x <- c(1, NA, 3, 4, 5)
  y <- c(2,  4, 5, 4, 5)
  expect_equal(
    cor(vecvec(x), y = y, use = "complete.obs"),
    cor(x, y = y, use = "complete.obs")
  )
})

test_that("cor() respects method = 'spearman'", {
  x <- c(1.0, 2.0, 3.0, 4.0, 5.0)
  y <- c(5.0, 6.0, 7.0, 8.0, 7.0)
  expect_equal(
    cor(vecvec(x), y = y, method = "spearman"),
    cor(x, y = y, method = "spearman")
  )
})

# unique() ---------------------------------------------------------------------

test_that("unique() returns unchanged vecvec when no duplicates", {
  vv <- vecvec(1:10)
  expect_identical(unique(vv), vv)
})

test_that("unique() removes duplicated elements", {
  vv <- c(vecvec(1:5), vecvec(3:7))
  expect_identical(
    as.integer(unique(vv)),
    1:7
  )
})

test_that("unique() on a zero-length vecvec returns zero-length vecvec", {
  vv <- vecvec()
  expect_identical(length(unique(vv)), 0L)
})

test_that("unique() on a single-element vecvec returns the same element", {
  vv <- vecvec(99L)
  expect_identical(unique(vv), vv)
})

# Complex group: Re(), Im(), Mod(), Arg(), Conj()

# Helpers
make_complex_vv <- function() vecvec(1:5 + 0i, complex(real = 0, imaginary = 1:3))
make_real_vv    <- function() vecvec(1:5, 6:10)

# --- Re() -------------------------------------------------------------------

test_that("Re() on a purely real vecvec returns its values unchanged", {
  x <- 1:5
  expect_equal(Re(vecvec(x)), as.double(x))
})

test_that("Re() extracts real parts from a complex vecvec", {
  z <- c(1 + 2i, 3 + 4i, 5 + 6i)
  expect_equal(Re(vecvec(z)), Re(z))
})

test_that("Re() on a multi-segment vecvec matches base Re()", {
  vv <- make_complex_vv()
  expect_equal(Re(vv), Re(unvecvec(vv)))
})

test_that("Re() result has correct length", {
  vv <- make_complex_vv()
  expect_equal(length(Re(vv)), length(vv))
})

# --- Im() -------------------------------------------------------------------

test_that("Im() on a purely real vecvec returns zeros", {
  expect_equal(Im(vecvec(1:5)), rep(0, 5))
})

test_that("Im() extracts imaginary parts from a complex vecvec", {
  z <- c(1 + 2i, 3 + 4i, 5 + 6i)
  expect_equal(Im(vecvec(z)), Im(z))
})

test_that("Im() on a multi-segment vecvec matches base Im()", {
  vv <- make_complex_vv()
  expect_equal(Im(vv), Im(unvecvec(vv)))
})

test_that("Im() result has correct length", {
  vv <- make_complex_vv()
  expect_equal(length(Im(vv)), length(vv))
})

# --- Mod() ------------------------------------------------------------------

test_that("Mod() on a purely real positive vecvec equals the values", {
  x <- 1:5
  expect_equal(Mod(vecvec(x)), as.double(x))
})

test_that("Mod() on a purely real negative vecvec returns absolute values", {
  x <- -5:-1
  expect_equal(Mod(vecvec(x)), abs(as.double(x)))
})

test_that("Mod() on a complex vecvec matches base Mod()", {
  z <- c(3 + 4i, 5 + 12i)
  expect_equal(Mod(vecvec(z)), Mod(z))
})

test_that("Mod() on a multi-segment complex vecvec matches base Mod()", {
  vv <- make_complex_vv()
  expect_equal(Mod(vv), Mod(unvecvec(vv)))
})

# --- Arg() ------------------------------------------------------------------

test_that("Arg() on a positive real vecvec returns zeros", {
  expect_equal(Arg(vecvec(1:5)), rep(0, 5))
})

test_that("Arg() on a purely imaginary vecvec returns pi/2", {
  z <- c(0 + 1i, 0 + 2i)
  expect_equal(Arg(vecvec(z)), Arg(z))
})

test_that("Arg() on a complex vecvec matches base Arg()", {
  z <- c(1 + 1i, -1 + 1i, -1 - 1i, 1 - 1i)
  expect_equal(Arg(vecvec(z)), Arg(z))
})

test_that("Arg() on a multi-segment vecvec matches base Arg()", {
  vv <- make_complex_vv()
  expect_equal(Arg(vv), Arg(unvecvec(vv)))
})

# --- Conj() -----------------------------------------------------------------

test_that("Conj() on a purely real vecvec returns it unchanged (as complex)", {
  z <- c(1 + 0i, 2 + 0i, 3 + 0i)
  expect_equal(Conj(vecvec(z)), Conj(z))
})

test_that("Conj() negates imaginary parts", {
  z <- c(1 + 2i, 3 + 4i, 5 + 6i)
  expect_equal(Conj(vecvec(z)), Conj(z))
})

test_that("Conj() on a multi-segment complex vecvec matches base Conj()", {
  vv <- make_complex_vv()
  expect_equal(Conj(vv), Conj(unvecvec(vv)))
})

test_that("Conj() result has correct length", {
  z <- c(1 + 2i, 3 + 4i)
  expect_equal(length(Conj(vecvec(z))), length(z))
})

test_that("Conj() applied twice returns the original values", {
  z <- c(1 + 2i, 3 - 4i, 0 + 5i)
  expect_equal(Conj(Conj(vecvec(z))), z)
})

test_that("format() on empty vecvec returns character(0)", {
  expect_identical(format(vecvec()), character(0))
})

test_that("format() on non-empty vecvec returns formatted strings", {
  expect_identical(format(vecvec(1:3, letters[1:3])), c("1", "2", "3", "a", "b", "c"))
})

test_that("print() on empty vecvec produces no output beyond the header", {
  expect_snapshot(print(vecvec()))
})

test_that("print() on non-empty vecvec produces expected output", {
  expect_snapshot(print(vecvec(1:3, letters[1:3])))
})

test_that("unvecvec", {
  expect_equal(
    unvecvec(x <- vecvec(1:5, pi, exp(1))),
    y <- c(1:5, pi, exp(1))
  )

  expect_equal(
    unvecvec(x[c(3, 7, 4, 5, 6, 2, 1)]),
    y[c(3, 7, 4, 5, 6, 2, 1)]
  )
})

test_that("unvecvec zero-length", {
  expect_equal(
    unvecvec(vecvec()),
    NULL
  )
  expect_equal(
    unvecvec(vecvec(numeric(0))),
    numeric(0)
  )
})

test_that("unvecvec missing values", {
  expect_equal(
    unvecvec(vecvec(c(1, NA, 3), c(NA, 5))),
    c(1, NA, 3, NA, 5)
  )
})


test_that("Replicating vectors", {
  rand <- rnorm(5, sd = 5)
  expect_equal(
    as.numeric(rep(vecvec(1:10, rand), 5)),
    rep(c(1:10, rand), 5)
  )
})

# Summary group: all(), any(), sum(), prod(), min(), max(), range()

# --- all() ------------------------------------------------------------------

test_that("all() returns TRUE when all elements are TRUE", {
  expect_true(all(vecvec(c(TRUE, TRUE), c(TRUE, TRUE))))
})

test_that("all() returns FALSE when any element is FALSE", {
  expect_false(all(vecvec(c(TRUE, FALSE), c(TRUE, TRUE))))
})

test_that("all() handles a single-segment vecvec", {
  expect_true(all(vecvec(c(TRUE, TRUE, TRUE))))
  expect_false(all(vecvec(c(TRUE, FALSE, TRUE))))
})

test_that("all() respects na.rm", {
  expect_true(all(vecvec(c(TRUE, NA), c(TRUE)), na.rm = TRUE))
  expect_false(all(vecvec(c(TRUE, FALSE), c(NA)), na.rm = TRUE))
  expect_identical(all(vecvec(c(TRUE, NA))), NA)
})

test_that("all() works across multiple vecvec arguments", {
  expect_true(all(vecvec(c(TRUE, TRUE)), vecvec(c(TRUE, TRUE))))
  expect_false(all(vecvec(c(TRUE, TRUE)), vecvec(c(TRUE, FALSE))))
})

test_that("all() works with mixed vecvec and plain vector arguments", {
  expect_true(all(vecvec(c(TRUE, TRUE)), c(TRUE, TRUE)))
  expect_false(all(vecvec(c(TRUE, TRUE)), c(TRUE, FALSE)))
})

# --- any() ------------------------------------------------------------------

test_that("any() returns TRUE when at least one element is TRUE", {
  expect_true(any(vecvec(c(FALSE, TRUE), c(FALSE))))
})

test_that("any() returns FALSE when all elements are FALSE", {
  expect_false(any(vecvec(c(FALSE, FALSE), c(FALSE, FALSE))))
})

test_that("any() respects na.rm", {
  expect_false(any(vecvec(c(FALSE, NA)), na.rm = TRUE))
  expect_identical(any(vecvec(c(FALSE, NA))), NA)
})

test_that("any() works across multiple vecvec arguments", {
  expect_false(any(vecvec(c(FALSE, FALSE)), vecvec(c(FALSE, FALSE))))
  expect_true(any(vecvec(c(FALSE, FALSE)), vecvec(c(FALSE, TRUE))))
})

# --- sum() ------------------------------------------------------------------

test_that("sum() of integer vecvec matches base sum()", {
  x <- 1:10
  expect_equal(sum(vecvec(x)), sum(x))
})

test_that("sum() of double vecvec matches base sum()", {
  x <- c(1.1, 2.2, 3.3)
  expect_equal(sum(vecvec(x)), sum(x))
})

test_that("sum() across multiple segments matches base sum()", {
  x <- 1:5
  y <- 6:10
  expect_equal(sum(vecvec(x, y)), sum(c(x, y)))
})

test_that("sum() respects na.rm", {
  x <- c(1, NA, 3)
  expect_equal(sum(vecvec(x), na.rm = TRUE), sum(x, na.rm = TRUE))
  expect_identical(sum(vecvec(x)), NA_real_)
})

test_that("sum() works across multiple vecvec arguments", {
  expect_equal(
    sum(vecvec(1:5), vecvec(6:10)),
    sum(1:10)
  )
})

test_that("sum() works with mixed vecvec and plain vector arguments", {
  expect_equal(sum(vecvec(1:5), 6:10), sum(1:10))
})

# --- prod() -----------------------------------------------------------------

test_that("prod() of integer vecvec matches base prod()", {
  x <- 1:5
  expect_equal(prod(vecvec(x)), prod(x))
})

test_that("prod() across multiple segments matches base prod()", {
  x <- c(1, 2, 3)
  y <- c(4, 5, 6)
  expect_equal(prod(vecvec(x, y)), prod(c(x, y)))
})

test_that("prod() respects na.rm", {
  x <- c(2, NA, 4)
  expect_equal(prod(vecvec(x), na.rm = TRUE), prod(x, na.rm = TRUE))
  expect_identical(prod(vecvec(x)), NA_real_)
})

# --- min() ------------------------------------------------------------------

test_that("min() of integer vecvec matches base min()", {
  expect_equal(min(vecvec(1:10)), 1L)
})

test_that("min() across multiple segments matches base min()", {
  expect_equal(min(vecvec(5:10, 1:4)), 1L)
})

test_that("min() of double vecvec matches base min()", {
  x <- c(3.5, 1.2, 4.8)
  expect_equal(min(vecvec(x)), min(x))
})

test_that("min() respects na.rm", {
  x <- c(5L, NA_integer_, 3L)
  expect_equal(min(vecvec(x), na.rm = TRUE), 3L)
  expect_identical(min(vecvec(x)), NA_integer_)
})

test_that("min() works across multiple vecvec arguments", {
  expect_equal(min(vecvec(5:10), vecvec(1:4)), 1L)
})

test_that("min() works with mixed vecvec and plain vector arguments", {
  expect_equal(min(vecvec(5:10), 1:4), 1L)
})

# --- max() ------------------------------------------------------------------

test_that("max() of integer vecvec matches base max()", {
  expect_equal(max(vecvec(1:10)), 10L)
})

test_that("max() across multiple segments matches base max()", {
  expect_equal(max(vecvec(1:5, 6:10)), 10L)
})

test_that("max() of double vecvec matches base max()", {
  x <- c(3.5, 1.2, 4.8)
  expect_equal(max(vecvec(x)), max(x))
})

test_that("max() respects na.rm", {
  x <- c(5L, NA_integer_, 3L)
  expect_equal(max(vecvec(x), na.rm = TRUE), 5L)
  expect_identical(max(vecvec(x)), NA_integer_)
})

test_that("max() works across multiple vecvec arguments", {
  expect_equal(max(vecvec(1:5), vecvec(6:10)), 10L)
})

test_that("max() works with mixed vecvec and plain vector arguments", {
  expect_equal(max(vecvec(1:5), 6:10), 10L)
})

# --- range() ----------------------------------------------------------------

test_that("range() of integer vecvec matches base range()", {
  expect_equal(range(vecvec(1:10)), c(1L, 10L))
})

test_that("range() across multiple segments matches base range()", {
  expect_equal(range(vecvec(5:10, 1:4)), c(1L, 10L))
})

test_that("range() of double vecvec matches base range()", {
  x <- c(3.5, 1.2, 4.8)
  expect_equal(range(vecvec(x)), range(x))
})

test_that("range() respects na.rm", {
  x <- c(3L, NA_integer_, 7L)
  expect_equal(range(vecvec(x), na.rm = TRUE), c(3L, 7L))
})

# Helpers ------------------------------------------------------------------

vv_chr  <- function() vecvec(letters)
vv_dbl  <- function() vecvec(as.double(1:26))
vv_int  <- function() vecvec(1:26)
vv_date <- function() vecvec(Sys.Date() + 0:25)
vv_mix  <- function() vecvec(letters, as.double(1:26), 1:26, Sys.Date() + 0:25)

# dim<- / array construction -----------------------------------------------

test_that("dim<- sets dim attribute on vecvec", {
  x <- vv_chr()
  dim(x) <- c(2L, 13L)
  expect_equal(dim(x), c(2L, 13L))
  expect_true(is_vecvec(x))
})

test_that("array() wraps a single-type vecvec preserving class and dim", {
  x <- array(vv_int(), dim = c(2L, 13L))
  expect_equal(dim(x), c(2L, 13L))
  expect_true(is_vecvec(x))
})

test_that("array() wraps a mixed vecvec into a 2D array", {
  x <- array(vv_mix(), dim = c(8L, 13L))
  expect_equal(dim(x), c(8L, 13L))
  expect_true(is_vecvec(x))
})

test_that("array() wraps a mixed vecvec into a 3D array", {
  x <- array(vv_mix(), dim = c(8L, 3L, 4L))
  expect_equal(dim(x), c(8L, 3L, 4L))
  expect_true(is_vecvec(x))
})

test_that("dim<- round-trips: setting then reading dim is consistent", {
  x <- vv_dbl()
  dim(x) <- c(2L, 13L)
  expect_equal(dim(x), c(2L, 13L))
  dim(x) <- NULL
  expect_null(dim(x))
})

test_that("length is unchanged after dim<- assignment", {
  x <- vv_chr()
  dim(x) <- c(2L, 13L)
  expect_equal(length(x), 26L)
})

test_that("is_vecvec is TRUE for dim-annotated vecvec", {
  x <- vv_mix()
  dim(x) <- c(8L, 13L)
  expect_true(is_vecvec(x))
})

test_that("dim<- with a single dimension (vector form) is valid", {
  x <- vv_int()
  dim(x) <- 26L
  expect_equal(dim(x), 26L)
  expect_true(is_vecvec(x))
})

# dimnames<- ---------------------------------------------------------------

test_that("dimnames<- assigns row and column names", {
  x <- array(vv_int(), dim = c(2L, 13L))
  rn <- paste0("r", 1:2)
  cn <- paste0("c", 1:13)
  dimnames(x) <- list(rn, cn)
  expect_equal(dimnames(x), list(rn, cn))
})

test_that("dimnames<- with NULL clears names", {
  x <- array(vv_int(), dim = c(2L, 13L))
  dimnames(x) <- list(paste0("r", 1:2), paste0("c", 1:13))
  dimnames(x) <- NULL
  expect_null(dimnames(x))
})

test_that("dimnames<- preserves vecvec class and length", {
  x <- array(vv_int(), dim = c(2L, 13L))
  dimnames(x) <- list(paste0("r", 1:2), paste0("c", 1:13))
  expect_true(is_vecvec(x))
  expect_equal(length(x), 26L)
})

test_that("dimnames<- works for 3D array", {
  x <- array(vv_int(), dim = c(2L, 3L, 4L))
  dn <- list(
    paste0("r", 1:2),
    paste0("c", 1:3),
    paste0("z", 1:4)
  )
  dimnames(x) <- dn
  expect_equal(dimnames(x), dn)
  expect_true(is_vecvec(x))
})

# as.matrix ----------------------------------------------------------------

test_that("as.matrix on a character vecvec gives a column matrix", {
  m <- as.matrix(vv_chr())
  expect_true(is_vecvec(m))
  expect_equal(dim(m), c(26L, 1L))
})

test_that("as.matrix on a double vecvec gives a column matrix", {
  m <- as.matrix(vv_dbl())
  expect_true(is_vecvec(m))
  expect_equal(dim(m), c(26L, 1L))
})

test_that("as.matrix preserves element values via [[", {
  m <- as.matrix(vv_chr())
  expect_equal(m[[1L]], "a")
  expect_equal(m[[26L]], "z")
})

test_that("as.matrix on a mixed vecvec produces correct dim", {
  m <- as.matrix(vv_mix())
  expect_true(is_vecvec(m))
  expect_equal(dim(m), c(104L, 1L))
})

test_that("as.matrix preserves Date elements via [[", {
  m <- as.matrix(vv_date())
  expect_s3_class(m[[1L]], "Date")
})

# t (transpose) ------------------------------------------------------------

test_that("t() transposes a column matrix to a row matrix", {
  m  <- as.matrix(vv_chr())
  mt <- t(m)
  expect_equal(dim(mt), c(1L, 26L))
  expect_true(is_vecvec(mt))
})

test_that("t() on a 2-row integer matrix transposes dim correctly", {
  x  <- array(vv_int(), dim = c(2L, 13L))
  xt <- t(x)
  expect_equal(dim(xt), c(13L, 2L))
  expect_true(is_vecvec(xt))
})

test_that("t() preserves element values in transposed positions (via format)", {
  x  <- array(vv_int(), dim = c(2L, 13L))
  xt <- t(x)
  f  <- format(xt)
  # column-major: original [1,1]=1, [2,1]=2; after t: [1,1]=1, [1,2]=2
  expect_equal(trimws(f[1L, 1L]), "1")
  expect_equal(trimws(f[1L, 2L]), "2")
  expect_equal(trimws(f[2L, 1L]), "3")
})

test_that("t(t(x)) round-trips dim to original", {
  x <- array(vv_int(), dim = c(2L, 13L))
  expect_equal(dim(t(t(x))), dim(x))
})

# flat [[]] indexing on dim-annotated vecvec --------------------------------

test_that("[[i]] on a dim-annotated single-type vecvec returns scalar", {
  x <- array(vv_int(), dim = c(2L, 13L))
  expect_equal(x[[1L]], 1L)
  expect_equal(x[[26L]], 26L)
})

test_that("[[i]] on a dim-annotated mixed vecvec returns correct type", {
  x <- array(vv_mix(), dim = c(8L, 13L))
  expect_equal(x[[1L]],  "a")
  expect_equal(x[[26L]], "z")
  expect_equal(x[[27L]], 1)         # first double
  expect_s3_class(x[[79L]], "Date") # first Date
})

test_that("[[i]] on a 1x1 array returns the single element", {
  x <- array(vecvec(42L), dim = c(1L, 1L))
  expect_equal(x[[1L]], 42L)
})

# [i] flat indexing preserves class and dim --------------------------------

# [, j] column subset (note: [i, ] and [i, j] currently error) -----------

test_that("[, j] on a dim-annotated vecvec returns a vecvec", {
  x <- array(vv_int(), dim = c(2L, 13L))
  col1 <- x[, 1L]
  expect_true(is_vecvec(col1))
})

test_that("[i, ] on a dim-annotated vecvec returns the correct row as a vecvec", {
  x <- array(vv_int(), dim = c(2L, 13L))
  row1 <- x[1L, ]
  expect_true(is_vecvec(row1))
  expect_equal(as.integer(row1), seq(1L, 25L, by = 2L))
})

test_that("[i, j] on a dim-annotated vecvec returns the correct scalar element", {
  x <- array(vv_int(), dim = c(2L, 13L))
  expect_equal(x[1L, 1L], vecvec(1L))
  expect_equal(x[2L, 1L], vecvec(2L))
  expect_equal(x[1L, 2L], vecvec(3L))
})

# format -------------------------------------------------------------------

test_that("format() on a 2D dim-annotated vecvec returns a character matrix", {
  x <- array(vv_int(), dim = c(2L, 13L))
  f <- format(x)
  expect_true(is.character(f))
  expect_equal(dim(f), c(2L, 13L))
})

test_that("format() on 3D mixed vecvec array returns a character array", {
  x <- array(vv_mix(), dim = c(8L, 3L, 4L))
  f <- format(x)
  expect_true(is.character(f))
  expect_equal(dim(f), c(8L, 3L, 4L))
})

test_that("format() values match underlying elements (character slot)", {
  x <- array(vv_chr(), dim = c(2L, 13L))
  f <- format(x)
  expect_equal(trimws(f[1L, 1L]), "a")
  expect_equal(trimws(f[2L, 1L]), "b")
  expect_equal(trimws(f[1L, 2L]), "c")
})

# as.vector ----------------------------------------------------------------

test_that("as.vector on a dim-annotated vecvec retains the dim attribute", {
  # Current behaviour: as.vector is identity for vecvec (dim preserved)
  x <- array(vv_chr(), dim = c(2L, 13L))
  v <- as.vector(x)
  expect_true(is_vecvec(v))
  expect_equal(length(v), 26L)
})

# as.character / as.double / as.integer coercion --------------------------

test_that("as.character on a dim-annotated character vecvec returns letters", {
  x  <- array(vv_chr(), dim = c(2L, 13L))
  ch <- as.character(x)
  expect_equal(ch, letters)
})

test_that("as.integer on a dim-annotated integer vecvec returns 1:26", {
  x <- array(vv_int(), dim = c(2L, 13L))
  expect_equal(as.integer(x), 1:26)
})

test_that("as.double on a dim-annotated double vecvec returns correct values", {
  x <- array(vv_dbl(), dim = c(2L, 13L))
  expect_equal(as.double(x), as.double(1:26))
})

# [i, j, ...] replacement -------------------------------------------------

test_that("[i, j]<- replaces a single element in a 2D integer vecvec", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[1L, 1L] <- 99L
  expect_equal(x[[1L]], 99L)
})

test_that("[i, j]<- replaces an element that is not the first, in a 2D integer vecvec", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[2L, 1L] <- 99L
  expect_equal(x[[2L]], 99L)
})

test_that("[i, j]<- preserves surrounding elements in a 2D array", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[1L, 1L] <- 99L
  expect_equal(x[[2L]], 2L)
  expect_equal(x[[3L]], 3L)
})

test_that("[i, j]<- preserves dim after replacement", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[1L, 2L] <- 99L
  expect_equal(dim(x), c(2L, 13L))
})

test_that("[i, j]<- preserves vecvec class after replacement", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[2L, 3L] <- 99L
  expect_true(is_vecvec(x))
})

test_that("[i, j]<- preserves length after replacement", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[1L, 1L] <- 99L
  expect_equal(length(x), 26L)
})

test_that("[, j]<- replaces an entire column in a 2D array", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[, 1L] <- c(91L, 92L)
  expect_equal(x[[1L]], 91L)
  expect_equal(x[[2L]], 92L)
  expect_equal(x[[3L]], 3L)  # column 2, row 1 unchanged
})

test_that("[i, ]<- replaces an entire row in a 2D array", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[1L, ] <- 0L
  # row 1: positions 1, 3, 5, ... (odd flat indices)
  expect_equal(x[[1L]], 0L)
  expect_equal(x[[3L]], 0L)
  expect_equal(x[[2L]], 2L)  # row 2, col 1 unchanged
})

test_that("[i, j]<- can replace with a different type (character into integer vecvec)", {
  x <- array(vv_int(), dim = c(2L, 13L))
  x[1L, 1L] <- "hello"
  expect_equal(x[[1L]], "hello")
  expect_equal(x[[2L]], 2L)
})

test_that("[i, j]<- can replace with a Date value", {
  x <- array(vv_int(), dim = c(2L, 13L))
  d <- Sys.Date()
  x[1L, 1L] <- d
  expect_s3_class(x[[1L]], "Date")
  expect_equal(x[[1L]], d)
})

test_that("[i, j]<- works on a 2D mixed vecvec", {
  x <- array(vv_mix(), dim = c(8L, 13L))
  x[1L, 1L] <- 999L
  expect_equal(x[[1L]], 999L)
  expect_equal(dim(x), c(8L, 13L))
  expect_true(is_vecvec(x))
})

test_that("[i, j, k]<- replaces a single element in a 3D integer vecvec", {
  x <- array(vv_int(), dim = c(2L, 3L, 4L))
  x[1L, 1L, 1L] <- 99L
  expect_equal(x[[1L]], 99L)
})

test_that("[i, j, k]<- replaces an element not at position 1 in a 3D array", {
  x <- array(vv_int(), dim = c(2L, 3L, 4L))
  # flat index for [2,3,4] in column-major: 2 + (3-1)*2 + (4-1)*6 = 2+4+18 = 24
  x[2L, 3L, 4L] <- 99L
  expect_equal(x[[24L]], 99L)
})

test_that("[i, j, k]<- preserves dim after replacement in a 3D array", {
  x <- array(vv_int(), dim = c(2L, 3L, 4L))
  x[1L, 2L, 3L] <- 99L
  expect_equal(dim(x), c(2L, 3L, 4L))
})

test_that("[i, j, k]<- preserves vecvec class and length in a 3D array", {
  x <- array(vv_int(), dim = c(2L, 3L, 4L))
  x[2L, 2L, 2L] <- 0L
  expect_true(is_vecvec(x))
  expect_equal(length(x), 24L)
})

test_that("[i, j, k]<- preserves surrounding elements in a 3D array", {
  x <- array(vv_int(), dim = c(2L, 3L, 4L))
  x[1L, 1L, 1L] <- 99L
  expect_equal(x[[2L]], 2L)
  expect_equal(x[[3L]], 3L)
})

test_that("[i, j, k]<- works on a 3D mixed vecvec", {
  x <- array(vv_mix(), dim = c(8L, 3L, 4L))
  x[1L, 1L, 1L] <- "replaced"
  expect_equal(x[[1L]], "replaced")
  expect_equal(dim(x), c(8L, 3L, 4L))
  expect_true(is_vecvec(x))
})

test_that("[i]<- on a 1D dim-annotated vecvec replaces the element and preserves dim", {
  x <- vv_int()
  dim(x) <- 26L
  x[3L] <- 99L
  expect_equal(x[[3L]], 99L)
  expect_equal(dim(x), 26L)
  expect_true(is_vecvec(x))
})

# Edge cases ---------------------------------------------------------------

test_that("zero-length vecvec with dim c(0, 0) is valid", {
  x <- vecvec()
  dim(x) <- c(0L, 0L)
  expect_equal(dim(x), c(0L, 0L))
  expect_equal(length(x), 0L)
  expect_true(is_vecvec(x))
})

test_that("1x1 array of vecvec has correct dim and element", {
  x <- array(vecvec(42L), dim = c(1L, 1L))
  expect_equal(dim(x), c(1L, 1L))
  expect_true(is_vecvec(x))
  expect_equal(x[[1L]], 42L)
})

test_that("3D array of mixed vecvec: Date elements accessible via [[", {
  x <- array(vv_mix(), dim = c(8L, 3L, 4L))
  # vv_mix: 26 chars + 26 doubles + 26 ints + 26 dates = 104 source elements;
  # array dim is 8×3×4 = 96 slots, so Date slot (starts at index 79) fills positions 79–96.
  expect_s3_class(x[[79L]], "Date")
  expect_s3_class(x[[96L]], "Date")
})

# is.finite() ----------------------------------------------------------------

test_that("is.finite() returns TRUE for vecvec with all finite values", {
  x <- vecvec(c(1, 2, 3))
  expect_equal(is.finite(x), c(TRUE, TRUE, TRUE))
})

test_that("is.finite() returns FALSE for Inf in a vecvec", {
  x <- vecvec(c(1, Inf, 3))
  expect_equal(is.finite(x), c(TRUE, FALSE, TRUE))
})

test_that("is.finite() returns FALSE for -Inf in a vecvec", {
  x <- vecvec(c(-Inf, 2, 3))
  expect_equal(is.finite(x), c(FALSE, TRUE, TRUE))
})

test_that("is.finite() returns FALSE for NaN in a vecvec", {
  x <- vecvec(c(1, NaN, 3))
  expect_equal(is.finite(x), c(TRUE, FALSE, TRUE))
})

test_that("is.finite() returns FALSE for NA in a vecvec", {
  x <- vecvec(c(1, NA, 3))
  expect_equal(is.finite(x), c(TRUE, FALSE, TRUE))
})

test_that("is.finite() works on a length-1 finite vecvec", {
  x <- vecvec(1.5)
  expect_true(is.finite(x))
})

test_that("is.finite() works on a length-1 Inf vecvec", {
  x <- vecvec(Inf)
  expect_false(is.finite(x))
})

# is.infinite() --------------------------------------------------------------

test_that("is.infinite() returns FALSE for vecvec with no infinite values", {
  x <- vecvec(c(1, 2, 3))
  expect_equal(is.infinite(x), c(FALSE, FALSE, FALSE))
})

test_that("is.infinite() detects Inf in a vecvec", {
  x <- vecvec(c(1, Inf, 3))
  expect_equal(is.infinite(x), c(FALSE, TRUE, FALSE))
})

test_that("is.infinite() detects -Inf in a vecvec", {
  x <- vecvec(c(-Inf, 2, 3))
  expect_equal(is.infinite(x), c(TRUE, FALSE, FALSE))
})

test_that("is.infinite() detects Inf and -Inf at the start and end of a vecvec", {
  x <- vecvec(c(Inf, 2, -Inf))
  expect_equal(is.infinite(x), c(TRUE, FALSE, TRUE))
})

test_that("is.infinite() returns FALSE for NaN in a vecvec", {
  x <- vecvec(c(1, NaN, 3))
  expect_equal(is.infinite(x), c(FALSE, FALSE, FALSE))
})

test_that("is.infinite() returns FALSE for NA in a vecvec", {
  x <- vecvec(c(1, NA, 3))
  expect_equal(is.infinite(x), c(FALSE, FALSE, FALSE))
})

test_that("is.infinite() works on a length-1 Inf vecvec", {
  x <- vecvec(Inf)
  expect_true(is.infinite(x))
})

# is.nan() -------------------------------------------------------------------

test_that("is.nan() returns FALSE for vecvec with no NaN values", {
  x <- vecvec(c(1, 2, 3))
  expect_equal(is.nan(x), c(FALSE, FALSE, FALSE))
})

test_that("is.nan() detects NaN in a vecvec", {
  x <- vecvec(c(1, NaN, 3))
  expect_equal(is.nan(x), c(FALSE, TRUE, FALSE))
})

test_that("is.nan() detects NaN at the start and end of a vecvec", {
  x <- vecvec(c(NaN, 2, NaN))
  expect_equal(is.nan(x), c(TRUE, FALSE, TRUE))
})

test_that("is.nan() returns all TRUE for an all-NaN vecvec", {
  x <- vecvec(c(NaN, NaN, NaN))
  expect_equal(is.nan(x), c(TRUE, TRUE, TRUE))
})

test_that("is.nan() returns FALSE for NA (not NaN) in a vecvec", {
  x <- vecvec(c(1, NA, 3))
  expect_equal(is.nan(x), c(FALSE, FALSE, FALSE))
})

test_that("is.nan() returns FALSE for Inf in a vecvec", {
  x <- vecvec(c(1, Inf, 3))
  expect_equal(is.nan(x), c(FALSE, FALSE, FALSE))
})

test_that("is.nan() works on a length-1 NaN vecvec", {
  x <- vecvec(NaN)
  expect_true(is.nan(x))
})

# is.na() --------------------------------------------------------------------

test_that("is.na() returns FALSE for vecvec with no NAs", {
  x <- vecvec(c(1, 2, 3))
  expect_equal(is.na(x), c(FALSE, FALSE, FALSE))
})

test_that("is.na() detects NA in a vecvec (#8)", {
  x <- vecvec(c(1, NA, 3))
  expect_equal(is.na(x), c(FALSE, TRUE, FALSE))
})

test_that("is.na() detects NAs at the start and end of a vecvec", {
  x <- vecvec(c(NA, 2, NA))
  expect_equal(is.na(x), c(TRUE, FALSE, TRUE))
})

test_that("is.na() returns all TRUE for an all-NA vecvec", {
  x <- vecvec(c(NA, NA, NA))
  expect_equal(is.na(x), c(TRUE, TRUE, TRUE))
})

test_that("is.na() works on a length-1 NA vecvec", {
  x <- vecvec(NA_real_)
  expect_true(is.na(x))
})

test_that("is.na() returns TRUE for NaN (NaN is also NA)", {
  x <- vecvec(c(1, NaN, 3))
  expect_equal(is.na(x), c(FALSE, TRUE, FALSE))
})

# anyNA() ----------------------------------------------------------------------

test_that("anyNA() returns FALSE for vecvec with no NAs", {
  vv <- vecvec(c(1, 2, 3))
  expect_false(anyNA(vv))
})

test_that("anyNA() returns TRUE when NA is present in a vecvec", {
  vv <- vecvec(c(1, NA, 3))
  expect_true(anyNA(vv))
})

test_that("anyNA() returns TRUE when NA is at the start of a vecvec", {
  vv <- vecvec(c(NA, 2, 3))
  expect_true(anyNA(vv))
})

test_that("anyNA() returns TRUE when NA is at the end of a vecvec", {
  vv <- vecvec(c(1, 2, NA))
  expect_true(anyNA(vv))
})

test_that("anyNA() returns TRUE for an all-NA vecvec", {
  vv <- vecvec(c(NA, NA, NA))
  expect_true(anyNA(vv))
})

test_that("anyNA() returns TRUE for NaN (NaN is also NA)", {
  vv <- vecvec(c(1, NaN, 3))
  expect_true(anyNA(vv))
})

test_that("anyNA() returns FALSE for Inf in a vecvec", {
  vv <- vecvec(c(1, Inf, 3))
  expect_false(anyNA(vv))
})

test_that("anyNA() returns FALSE for -Inf in a vecvec", {
  vv <- vecvec(c(-Inf, 2, 3))
  expect_false(anyNA(vv))
})

test_that("anyNA() returns FALSE on a zero-length vecvec", {
  expect_false(anyNA(vecvec()))
})

test_that("anyNA() returns FALSE on a single-element finite vecvec", {
  expect_false(anyNA(vecvec(1.5)))
})

test_that("anyNA() returns TRUE on a single-element NA vecvec", {
  expect_true(anyNA(vecvec(NA_real_)))
})

test_that("anyNA() returns TRUE when NA appears across multiple sub-vectors", {
  vv <- c(vecvec(1:3), vecvec(c(4, NA, 6)))
  expect_true(anyNA(vv))
})

test_that("anyNA() returns FALSE when no NA across multiple sub-vectors", {
  vv <- c(vecvec(1:3), vecvec(4:6))
  expect_false(anyNA(vv))
})

test_that("anyNA() returns TRUE when NA is in the first of multiple sub-vectors", {
  vv <- c(vecvec(c(1, NA, 3)), vecvec(4:6))
  expect_true(anyNA(vv))
})

# duplicated() ------------------------------------------------------------------

test_that("duplicated() returns all FALSE when no duplicates", {
  vv <- vecvec(1:5, 6:10)
  expect_identical(
    duplicated(vv),
    rep(FALSE, 10)
  )
})

test_that("duplicated() detects duplicates within the same sub-vector type", {
  vv <- c(vecvec(1:5), vecvec(3:7))
  plain <- c(1:5, 3:7)
  expect_identical(
    duplicated(vv),
    duplicated(plain)
  )
})

test_that("duplicated() detects duplicates across sub-vector types", {
  vv <- c(vecvec(1L, 2L, 3L), vecvec(3L))
  expect_identical(
    duplicated(vv),
    c(FALSE, FALSE, FALSE, TRUE)
  )
})

test_that("duplicated() fromLast marks first occurrence as duplicate", {
  vv <- c(vecvec(1:5), vecvec(3:7))
  plain <- c(1:5, 3:7)
  expect_identical(
    duplicated(vv, fromLast = TRUE),
    duplicated(plain, fromLast = TRUE)
  )
})

test_that("duplicated() on a zero-length vecvec returns logical(0)", {
  expect_identical(duplicated(vecvec()), logical(0))
})

test_that("duplicated() on a single-element vecvec returns FALSE", {
  expect_identical(duplicated(vecvec(42L)), FALSE)
})

# anyDuplicated() --------------------------------------------------------------

test_that("anyDuplicated() returns 0L when no duplicates", {
  vv <- vecvec(1:10)
  expect_identical(anyDuplicated(vv), 0L)
})

test_that("anyDuplicated() returns index of first duplicate", {
  vv <- c(vecvec(1:10, 3, 1L))
  expect_identical(anyDuplicated(vv), 12L)
})

test_that("anyDuplicated() fromLast returns index of last duplicate", {
  vv <- c(vecvec(1:5, "a", 3:7))
  expect_identical(
    anyDuplicated(vv, fromLast = TRUE),
    5L
  )
})

test_that("anyDuplicated() on a zero-length vecvec returns 0L", {
  expect_identical(anyDuplicated(vecvec()), 0L)
})

test_that("anyDuplicated() on a single-element vecvec returns 0L", {
  expect_identical(anyDuplicated(vecvec(1L)), 0L)
})

# `[<-` -------------------------------------------------------------------

test_that("`[<-` replaces a single element with a plain vector", {
  x <- vecvec(letters, 1:26)
  x[1] <- 99L
  expect_equal(x[[1]], 99L)
})

test_that("`[<-` replaces multiple elements with a plain vector", {
  x <- vecvec(1:10)
  x[1:3] <- c(11L, 12L, 13L)
  expect_equal(as.integer(x[1:3]), c(11L, 12L, 13L))
})

test_that("`[<-` recycles scalar value across selected indices", {
  x <- vecvec(1:5)
  x[1:3] <- 0L
  expect_equal(as.integer(x[1:3]), c(0L, 0L, 0L))
})

test_that("`[<-` preserves unmodified elements", {
  x <- vecvec(1:5)
  x[1] <- 99L
  expect_equal(as.integer(x[2:5]), 2:5)
})

test_that("`[<-` replaces elements across type boundaries", {
  x <- vecvec(letters, 1:26)
  x[25:28] <- 0L
  expect_equal(as.integer(x[25:28]), rep(0L, 4L))
})

test_that("`[<-` assigns a vecvec value", {
  x <- vecvec(letters, 1:26)
  x[1:3] <- vecvec(c(TRUE, FALSE, TRUE))
  expect_equal(as.logical(x[1:3]), c(TRUE, FALSE, TRUE))
})

test_that("`[<-` replaces with a Date value", {
  x <- vecvec(1:5)
  x[2:3] <- Sys.Date()
  expect_s3_class(x[[2]], "Date")
  expect_s3_class(x[[3]], "Date")
  expect_equal(x[[2]], x[[3]])
})

test_that("`[<-` keeps vecvec class after assignment", {
  x <- vecvec(1:5)
  x[1] <- 99L
  expect_true(is_vecvec(x))
})

test_that("`[<-` preserves length after single-element replacement", {
  x <- vecvec(letters)
  x[1] <- "z"
  expect_equal(length(x), 26L)
})

test_that("`[<-` preserves length after multi-element replacement", {
  x <- vecvec(1:10, letters)
  x[3:6] <- 0L
  expect_equal(length(x), 36L)
})

test_that("`[<-` replaces last element correctly", {
  x <- vecvec(1:5)
  x[5] <- 99L
  expect_equal(as.integer(x[5]), 99L)
  expect_equal(as.integer(x[1:4]), 1:4)
})

test_that("`[<-` replaces first element correctly", {
  x <- vecvec(1:5)
  x[1] <- 99L
  expect_equal(as.integer(x[1]), 99L)
  expect_equal(as.integer(x[2:5]), 2:5)
})

test_that("`[<-` round-trips through unvecvec after replacement", {
  x <- vecvec(1:5)
  x[3] <- 99L
  expected <- c(1L, 2L, 99L, 4L, 5L)
  expect_equal(unvecvec(x, ptype = integer()), expected)
})

# `[[<-` -------------------------------------------------------------------

test_that("`[[<-` replaces a single element", {
  x <- vecvec(letters)
  x[[1]] <- "z"
  expect_equal(x[[1]], "z")
})

test_that("`[[<-` replaces middle element", {
  x <- vecvec(1:5)
  x[[3]] <- 99L
  expect_equal(x[[3]], 99L)
})

test_that("`[[<-` preserves surrounding elements", {
  x <- vecvec(1:5)
  x[[3]] <- 99L
  expect_equal(as.integer(x[c(1, 2, 4, 5)]), c(1L, 2L, 4L, 5L))
})

test_that("`[[<-` preserves length", {
  x <- vecvec(letters)
  x[[1]] <- "z"
  expect_equal(length(x), 26L)
})

test_that("`[[<-` preserves vecvec class", {
  x <- vecvec(1:5)
  x[[1]] <- 99L
  expect_true(is_vecvec(x))
})

test_that("`[[<-` errors when multiple indices given", {
  x <- vecvec(1:5)
  expect_error(
    x[[1:2]] <- 99L,
    "attempt to replace multiple elements"
  )
})

test_that("`[[<-` works across type boundary", {
  x <- vecvec(letters, 1:26)
  x[[27]] <- 999L
  expect_equal(x[[27]], 999L)
})

# `is.na<-` ----------------------------------------------------------------

test_that("`is.na<-` sets a single element to NA by integer index", {
  x <- vecvec(1:5)
  is.na(x) <- 3L
  expect_true(is.na(x[[3]]))
})

test_that("`is.na<-` sets multiple elements to NA by integer indices", {
  x <- vecvec(1:5)
  is.na(x) <- c(1L, 3L, 5L)
  expect_equal(is.na(unvecvec(x, ptype = integer())), c(TRUE, FALSE, TRUE, FALSE, TRUE))
})

test_that("`is.na<-` sets elements to NA by logical vector", {
  x <- vecvec(1:5)
  is.na(x) <- c(TRUE, FALSE, TRUE, FALSE, FALSE)
  result <- is.na(unvecvec(x, ptype = integer()))
  expect_equal(result, c(TRUE, FALSE, TRUE, FALSE, FALSE))
})

test_that("`is.na<-` preserves non-NA elements", {
  x <- vecvec(1:5)
  is.na(x) <- 1L
  expect_equal(as.integer(x[2:5]), 2:5)
})

test_that("`is.na<-` preserves length", {
  x <- vecvec(1:5)
  is.na(x) <- 2L
  expect_equal(length(x), 5L)
})

test_that("`is.na<-` preserves vecvec class", {
  x <- vecvec(1:5)
  is.na(x) <- 1L
  expect_true(is_vecvec(x))
})

test_that("`is.na<-` errors on logical subscript of wrong length", {
  x <- vecvec(1:5)
  expect_error(
    { is.na(x) <- c(TRUE, FALSE) },
    "logical subscript has wrong length"
  )
})

test_that("`is.na<-` sets NA across type boundaries in a mixed vecvec", {
  x <- vecvec(letters, 1:5)
  is.na(x) <- c(25L, 26L, 27L)
  expect_true(is.na(x[[25]]))
  expect_true(is.na(x[[26]]))
  expect_true(is.na(x[[27]]))
})

test_that("`is.na<-` sets all elements NA with a fully-TRUE logical mask", {
  x <- vecvec(1:5)
  is.na(x) <- rep(TRUE, 5)
  expect_true(all(is.na(unvecvec(x, ptype = integer()))))
})

test_that("`is.na<-` with fully-FALSE mask leaves no NAs", {
  x <- vecvec(1:5)
  is.na(x) <- rep(FALSE, 5)
  expect_false(anyNA(unvecvec(x, ptype = integer())))
})

test_that("`is.na<-` produces NA of the correct type for character slot", {
  x <- vecvec(letters)
  is.na(x) <- 1L
  expect_true(is.na(x[[1]]))
  expect_true(is.character(x[[1]]))
})

test_that("`is.na<-` produces NA of the correct type for numeric slot", {
  x <- vecvec(rnorm(5))
  is.na(x) <- 2L
  expect_true(is.na(x[[2]]))
  expect_true(is.numeric(x[[2]]))
})

# vec_c -----------------------------------------------------------------

test_that("vctrs::vec_c", {
  expect_equal(
    vecvec(letters, 1:10),
    vec_c(vecvec(letters), vecvec(1:10))
  )
  expect_equal(
    vecvec(letters, 1:10),
    vec_c(vecvec(letters), 1:10)
  )
  expect_equal(
    vecvec(letters, 1:10),
    vec_c(letters, vecvec(1:10))
  )
})

# vec_init --------------------------------------------------------------

test_that("vctrs::vec_init", {
  vec <- vec_init(class_vecvec(), 5)
  expect_all_true(is.na(vec))
  expect_equal(
    as.numeric(vec),
    rep(NA_real_, 5)
  )
})

# vec_rep --------------------------------------------------------------

test_that("vctrs::vec_rep", {
  rand <- rnorm(5, sd = 5)
  expect_equal(
    as.numeric(vec_rep(vecvec(1:10, rand), 5)),
    rep(c(1:10, rand), 5)
  )
})

# vec_size ----------------------------------------------------------------

test_that("vctrs::vec_size", {
  expect_equal(vec_size(vecvec(1:5)), 5L)
  expect_equal(vec_size(vecvec(letters, 1:10)), 36L)
  expect_equal(vec_size(vecvec()), 0L)
  expect_equal(vec_size(vecvec(integer(0))), 0L)
})

# vec_slice ---------------------------------------------------------------

test_that("vctrs::vec_slice basic", {
  vv <- vecvec(1:5, 6:10)
  expect_equal(as.integer(vec_slice(vv, 1:5)), 1:5)
  expect_equal(as.integer(vec_slice(vv, 6:10)), 6:10)
  expect_equal(as.integer(vec_slice(vv, c(1L, 10L))), c(1L, 10L))
})

test_that("vctrs::vec_slice with mixed types", {
  vv <- vecvec(1:5, letters)
  expect_equal(as.integer(vec_slice(vv, 1:5)), 1:5)
  expect_equal(as.character(vec_slice(vv, 6:10)), letters[1:5])
})

test_that("vctrs::vec_slice out-of-order indices", {
  vv <- vecvec(1:5, 6:10)
  idx <- c(10L, 3L, 7L, 1L)
  expect_equal(as.integer(vec_slice(vv, idx)), c(10L, 3L, 7L, 1L))
})

test_that("vctrs::vec_slice zero-length result", {
  vv <- vecvec(1:5)
  expect_equal(vec_size(vec_slice(vv, integer(0))), 0L)
})

test_that("vctrs::vec_slice preserves vecvec class", {
  vv <- vecvec(1:5, letters)
  expect_true(is_vecvec(vec_slice(vv, 1:3)))
})

# proxy / restore roundtrip -----------------------------------------------

test_that("vctrs proxy/restore roundtrip is lossless", {
  vv <- vecvec(1:5, letters)
  expect_equal(vec_restore(vec_proxy(vv), vv), vv)
})

test_that("vctrs proxy/restore roundtrip: numeric only", {
  vv <- vecvec(rnorm(10))
  expect_equal(vec_restore(vec_proxy(vv), vv), vv)
})

test_that("vctrs proxy/restore roundtrip: zero-length", {
  vv <- class_vecvec()
  expect_equal(vec_restore(vec_proxy(vv), vv), vv)
})

# vec_ptype2 --------------------------------------------------------------

test_that("vctrs::vec_ptype2 vecvec x vecvec", {
  a <- vecvec(1:3)
  b <- vecvec(letters)
  pt <- vec_ptype2(a, b)
  expect_true(is_vecvec(pt))
  expect_equal(vec_size(pt), 0L)
})

test_that("vctrs::vec_ptype2 is commutative", {
  a <- vecvec(1:3)
  b <- vecvec(letters)
  expect_equal(vec_ptype2(a, b), vec_ptype2(b, a))
})

# vec_cast ----------------------------------------------------------------

# vecvec -> vecvec
test_that("vctrs::vec_cast vecvec to vecvec is identity", {
  vv <- vecvec(1:5, letters)
  expect_equal(vec_cast(vv, class_vecvec()), vv)
})

test_that("vctrs::vec_cast vecvec<integer> to vecvec<integer>", {
  vv <- vecvec(1L, 5L, 10L)
  result <- vec_cast(vv, vecvec(0L))
  expect_true(is_vecvec(result))
  expect_equal(as.integer(result), c(1L, 5L, 10L))
})

test_that("vctrs::vec_cast vecvec<character> to vecvec<character>", {
  vv <- vecvec(letters[1:3])
  result <- vec_cast(vv, vecvec(""))
  expect_true(is_vecvec(result))
  expect_equal(as.character(result), letters[1:3])
})

# plain type -> vecvec
test_that("vctrs::vec_cast plain vector to vecvec", {
  result <- vec_cast(1:5, class_vecvec())
  expect_true(is_vecvec(result))
  expect_equal(as.integer(result), 1:5)
})

test_that("vctrs::vec_cast integer to vecvec<integer>", {
  result <- vec_cast(10L, vecvec(1L))
  expect_true(is_vecvec(result))
  expect_equal(as.integer(result), 10L)
})

test_that("vctrs::vec_cast double to vecvec<double>", {
  result <- vec_cast(3.14, vecvec(1.0))
  expect_true(is_vecvec(result))
  expect_equal(as.double(result), 3.14)
})

test_that("vctrs::vec_cast character to vecvec<character>", {
  result <- vec_cast("hello", vecvec("world"))
  expect_true(is_vecvec(result))
  expect_equal(as.character(result), "hello")
})

test_that("vctrs::vec_cast logical to vecvec<logical>", {
  result <- vec_cast(c(TRUE, FALSE, NA), vecvec(TRUE))
  expect_true(is_vecvec(result))
  expect_equal(as.logical(result), c(TRUE, FALSE, NA))
})

# vecvec -> plain type
test_that("vctrs::vec_cast vecvec to integer", {
  vv <- vecvec(1L, 2L, 3L)
  result <- vec_cast(vv, integer())
  expect_equal(result, 1:3)
})

test_that("vctrs::vec_cast vecvec to double", {
  vv <- vecvec(c(1.5, 2.5, 3.5))
  result <- vec_cast(vv, double())
  expect_equal(result, c(1.5, 2.5, 3.5))
})

test_that("vctrs::vec_cast vecvec to character", {
  vv <- vecvec(c("a", "b", "c"))
  result <- vec_cast(vv, character())
  expect_equal(result, c("a", "b", "c"))
})

test_that("vctrs::vec_cast vecvec to logical", {
  vv <- vecvec(c(TRUE, FALSE))
  result <- vec_cast(vv, logical())
  expect_equal(result, c(TRUE, FALSE))
})

# zero-length edge cases
test_that("vctrs::vec_cast empty vecvec to vecvec is zero-length", {
  result <- vec_cast(class_vecvec(), class_vecvec())
  expect_true(is_vecvec(result))
  expect_equal(vec_size(result), 0L)
})

test_that("vctrs::vec_cast empty vector to vecvec is zero-length", {
  result <- vec_cast(integer(0), class_vecvec())
  expect_true(is_vecvec(result))
  expect_equal(vec_size(result), 0L)
})

# ptype labels ------------------------------------------------------------

test_that("vctrs::vec_ptype_abbr for mixed vecvec", {
  vv <- vecvec(1:3, letters)
  expect_equal(vec_ptype_abbr(vv), "vecvec")
})

test_that("vctrs::vec_ptype_full for mixed vecvec", {
  vv <- vecvec(1:3, letters)
  expect_equal(vec_ptype_full(vv), "vecvec")
})

test_that("vctrs::vec_ptype_abbr for single-type vecvec", {
  vv <- vecvec(c(1.5, 2.5))
  expect_equal(vec_ptype_abbr(vv), paste0(vec_ptype_abbr(numeric()), "*"))
})

test_that("vctrs::vec_ptype_full for single-type vecvec", {
  vv <- vecvec(c(1.5, 2.5))
  expect_equal(vec_ptype_full(vv), paste0(vec_ptype_full(numeric()), "*"))
})

# vec_detect_complete -----------------------------------------------------

test_that("vctrs::vec_detect_complete: no NAs", {
  vv <- vecvec(1:5)
  expect_true(all(vec_detect_complete(vv)))
})

test_that("vctrs::vec_detect_complete: with NAs", {
  vv <- vecvec(c(1, NA, 3))
  result <- vec_detect_complete(vv)
  expect_equal(result, c(TRUE, FALSE, TRUE))
})

test_that("vctrs::vec_detect_complete: all NAs", {
  vv <- vec_init(class_vecvec(), 3L)
  expect_true(all(!vec_detect_complete(vv)))
})

# vec_unique --------------------------------------------------------------

test_that("vctrs::vec_unique removes duplicates", {
  vv <- vec_c(vecvec(1:3), vecvec(1:3))
  u <- vec_unique(vv)
  expect_equal(vec_size(u), 3L)
  expect_equal(as.integer(u), 1:3)
})

test_that("vctrs::vec_unique on already-unique vecvec", {
  vv <- vecvec(1:5)
  expect_equal(vec_size(vec_unique(vv)), 5L)
})

# vec_order / vec_sort_indices --------------------------------------------

test_that("vctrs::vec_order ascending", {
  vv <- vecvec(c(3L, 1L, 2L))
  expect_equal(vec_order(vv), c(2L, 3L, 1L))
})

test_that("vctrs::vec_order descending", {
  vv <- vecvec(c(3L, 1L, 2L))
  expect_equal(vec_order(vv, direction = "desc"), c(1L, 3L, 2L))
})

test_that("vctrs::vec_sort ascending", {
  vv <- vecvec(c(3L, 1L, 4L, 1L, 5L))
  expect_equal(as.integer(vec_sort(vv)), c(1L, 1L, 3L, 4L, 5L))
})

method(is.finite, class_vecvec) <- vecvec_apply_fn(is.finite, ptype = logical())
method(is.infinite, class_vecvec) <- vecvec_apply_fn(is.infinite, ptype = logical())
method(is.nan, class_vecvec) <- vecvec_apply_fn(is.nan, ptype = logical())

method(is.na, class_vecvec) <- function(x) {
  # Missing values in vecvec indices or values are both considered NA.
  is.na(S7_data(x)) | unvecvec(vecvec_apply(x, is.na), ptype = logical())
}
method(anyNA, class_vecvec) <- function(x, recursive = FALSE) {
  if (anyNA(S7_data(x))) return(TRUE)
  
  for (v in x@x) {
    if (anyNA(v, recursive = recursive)) return(TRUE)
  }

  FALSE
}
method(na.fail, class_vecvec) <- function(object, ...) {
  if (anyNA(object)) {
    stop("missing values in object of class 'vecvec'", call. = FALSE)
  }
  object
}
na.drop <- function(object, class = NULL,...) {
  pos <- which(is.na(object))
  object <- object[-pos]
  
  # Add na.action attributes
  class(object) <- c(class, class(object))
  attr(object, "na.action") <- pos
  object
}
method(na.omit, class_vecvec) <- function(object, ...) na.drop(object, class = "omit", ...)
method(na.exclude, class_vecvec) <- function(object, ...) na.drop(object, class = "exclude", ...)

method(duplicated, class_vecvec) <- function(x, incomparables = FALSE, ...) {
  # Special case for empty vecvec vectors
  if (length(x@x) == 0L) {
    return(duplicated(S7_data(x), incomparables, ...))
  }

  # Find common vector types
  ptypes <- lapply(x@x, `[`, 0L)
  loc <- lapply(
    unique(ptypes),
    function(k) which(vapply(ptypes, identical, logical(1), k))
  )

  # Identify duplicated values within common vector types
  dup <- lapply(loc, function(i) {
    # Compute duplicates on a single vector
    vec <- unlist(x@x[i], recursive = FALSE)
    res <- duplicated(vec, incomparables = incomparables, ...)

    # Restructure result into list of vectors
    idx <- c(0L, cumsum(lengths(x@x[i])))
    len <- length(i)
    out <- vector("list", len)
    for (i in seq_len(len)) {
      out[[i]] <- res[seq(idx[i] + 1L, idx[i + 1])]
    }
    out
  })

  x@x[unlist(loc, recursive = FALSE)] <- unlist(dup, recursive = FALSE)
  unvecvec(x, ptype = logical())
}

method(anyDuplicated, class_vecvec) <- function(x, incomparables = FALSE, ...) {
  # Find common vector types
  ptypes <- lapply(x@x, `[`, 0L)
  loc <- lapply(
    unique(ptypes),
    function(k) which(vapply(ptypes, identical, logical(1), k))
  )

  # Search for any duplicated values within common vector types
  for (i in seq_along(loc)) {
    # Compute duplicates on a single vector
    idx <- loc[[i]]
    vec <- unlist(x@x[idx], recursive = FALSE)
    dup <- anyDuplicated(vec, incomparables = incomparables, ...)
    if (dup > 0L) {
      # Find the actual index of the duplicated value
      len <- c(0L, cumsum(lengths(x@x[idx[-length(idx)]])))
      pos <- findInterval(dup, len, left.open = TRUE)

      # Position on the original vector is the duplicated index minus the offset
      # of the current vector plus the offset of all previous vectors
      return(dup - len[pos] + sum(lengths(x@x[seq_len(loc[[i]][[pos]] - 1L)])))
    }
  }

  0L
}

register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }

  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}

#' Register methods for a vecvec subclass
#'
#' Call `vecvec_register()` inside your package's `.onLoad()` function to
#' register the S3 methods that allow a custom
#' [vecvec][vecvec::class_vecvec] subclass to participate in the
#' [vctrs](https://vctrs.r-lib.org/) vector system.
#'
#' Specifically, `vecvec_register()` registers:
#' - `vec_cast.<class>.*` methods, so that other types can be cast *to* your
#'   vecvec subclass.
#' - `vec_cast.*.<class>` methods, so that your vecvec subclass can be cast
#'   *from* other types.
#'
#' @param x An S7 class object that extends [class_vecvec]. Typically the
#'   class object created by [S7::new_class()] in your package, e.g.
#'   `class_myvec`.
#'
#' @return `NULL`, invisibly. Called for its side effects.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # In your package, define a vecvec subclass:
#' class_myvec <- S7::new_class(
#'   "myvec",
#'   parent = vecvec::class_vecvec
#' )
#'
#' # Then register it in .onLoad():
#' .onLoad <- function(libname, pkgname) {
#'   S7::methods_register()
#'   vecvec::vecvec_register(class_myvec)
#' }
#' }
vecvec_register <- function(x) {
  vctrs_exports <- getNamespaceExports(asNamespace("vctrs"))

  if (!S7_inherits(x(), class_vecvec)) {
    stop("`x` must be an S7 object extending `vecvec`", call. = FALSE)
  }

  pkg <- attr(x, "package", exact = TRUE)
  nm <- attr(x, "name", exact = TRUE)
  cls <- paste0(pkg, "::", nm)


  # --------------------------------------------------------
  # vctrs::vec_cast() methods
  # --------------------------------------------------------
  vec_cast_generics <- vctrs_exports[startsWith(vctrs_exports, "vec_cast.")]

  # Register vec_cast.*.<class> methods
  lapply(
    vec_cast_generics,
    register_s3_method,
    pkg = "vctrs",
    class = cls,
    fun = vec_cast_from_vecvec
  )
  # Register vec_cast.<class>.* methods
  lapply(
    sub("^vec_cast", cls, vec_cast_generics),
    register_s3_method,
    pkg = "vctrs",
    generic = "vec_cast",
    fun = vec_cast_to_vecvec
  )

  # --------------------------------------------------------
  # vctrs::vec_ptype2() methods
  # --------------------------------------------------------
  vec_ptype2_generics <- vctrs_exports[startsWith(vctrs_exports, "vec_ptype2.")]

  # Register vec_ptype2.*.<class> methods
  lapply(
    vec_ptype2_generics,
    register_s3_method,
    pkg = "vctrs",
    class = cls,
    fun = vec_ptype2_vecvec
  )
  # Register vec_ptype2.<class>.* methods
  lapply(
    sub("^vec_ptype2", cls, vec_ptype2_generics),
    register_s3_method,
    pkg = "vctrs",
    generic = "vec_ptype2",
    fun = vec_ptype2_vecvec
  )

  invisible(NULL)
}

#' S7 class for vecvec
#'
#' `class_vecvec()` constructs a new vecvec object from a list of vectors. It is
#' designed to be performant with minimal checks on the inputs. Direct use of
#' `class_vecvec()` is useful for developers, but for users it is recommended to
#' use [vecvec()] to create vecvec objects.
#' 
#' @param x An unnamed list of arbitrary vectors. The vectors can be of 
#'   different types and lengths, but they must all be vectors (according to
#'   [vctrs::vec_is]). Adjacent vectors of the same type will be combined 
#'   together into a single vector.
#' @param i A vector of integers specifying the location of each element in `x` 
#'   as if they were combined in order. The values in `i` must be between 1 and
#'   the total number of elements across all vectors in `x`, and can contain 
#'   duplicates. If not provided, it defaults to a sequence from 1 to the total 
#'   number of elements across all vectors in `x`.
#' 
#' @return A vector of vectors of S7 object of class `class_vecvec`.
#'
#' @examples
#' # Create a vecvec prototype
#' class_vecvec()
#'
#' # Construct a vecvec from a list of vectors
#' class_vecvec(list(letters, rnorm(10)))
#'
#' # Fully specify a vecvec with locations
#' class_vecvec(
#'   x = list(letters, rnorm(10)),
#'   i = c(1:3, 27:31, 26:4, 32:36)
#' )
#'
#' @export
class_vecvec <- new_class(
  "vecvec",
  parent = class_integer,
  properties = list(
    x = class_list
  ),
  constructor = function(x = list(), i = seq_len(sum(lengths(x)))) {
    new_object(i, x = vecvec_flatten_adj(x))
  },
  validator = function(self) {
    vec <- vapply(self@x, vec_is, logical(1L))
    if (!all(vec)) {
      "@x must be a list of primitive vectors"
    }
  }
)

# Alternative candidate implementation - indices in base object.
# class_vecvec <- S7::new_class(
#   "vecvec",
#   properties = list(
#     x = S7::class_list
#   ),
#   constructor = function(.data = seq_len(sum(lengths(x))), x = list()) {
#     .data
#     x
#     S7::new_object(class_integer, .data = .data, x = x)
#   },
#   validator = function(self) {
#     vec <- vapply(self@x, is.vector, logical(1L))
#     if (!all(vec)) {
#       "@x must be a list of vectors"
#     }
#   }
# )

# print() on empty vecvec produces no output beyond the header

    Code
      print(vecvec())
    Output
      <vecvec[0]>

# print() on non-empty vecvec produces expected output

    Code
      print(vecvec(1:3, letters[1:3]))
    Output
      <vecvec[6]>
      [1] 1 2 3 a b c

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vecvec.R
\name{unvecvec}
\alias{unvecvec}
\title{Coerce a vector of vectors to a single typed vector}
\usage{
unvecvec(x, ptype = NULL)
}
\arguments{
\item{x}{A \code{vecvec} object.}

\item{ptype}{A prototype specifying the desired output type, e.g.
\code{character()} or \code{numeric()}. If \code{NULL} (the default), the common type is
inferred from the elements of \code{x} using \code{\link[vctrs:vec_ptype_common]{vctrs::vec_ptype_common()}},
falling back to \code{logical()} when no common type can be determined. Passing
an explicit \code{ptype} is useful when you need a guaranteed output type
regardless of what \code{x} contains, or when automatic inference would pick an
undesirable type.}
}
\value{
A vector of length \code{length(x)} and type \code{ptype} (or the inferred
common type when \code{ptype = NULL}). Positions corresponding to \code{NA} indices
in the underlying \code{vecvec} structure are filled with \code{NA}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Flattens a \code{vecvec} into a plain R vector by casting all elements to a
common type. This is the inverse of \code{\link[=vecvec]{vecvec()}}.

Type resolution follows vctrs coercion rules: when \code{ptype} is \code{NULL} the
common type is determined automatically from the slots of \code{x} via
\code{\link[vctrs:vec_ptype_common]{vctrs::vec_ptype_common()}}. If no common type can be found (e.g. all slots
are empty), the result falls back to \code{logical()}.
}
\examples{
vv <- vecvec(1:3, c(4.5, 5.5))

# Automatic type inference: integer + double -> double
unvecvec(vv)

# Force a specific output type
unvecvec(vv, ptype = character())

}
\seealso{
\code{\link[=vecvec]{vecvec()}} to create a \code{vecvec}; \code{\link[vctrs:vec_ptype_common]{vctrs::vec_ptype_common()}} for
the type inference rules; \code{\link[vctrs:vec_cast]{vctrs::vec_cast()}} for the casting rules.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vecvec-package.R
\docType{package}
\name{vecvec-package}
\alias{vecvec-package}
\title{vecvec: Construct Mixed Type Data Structures with Vectors of Vectors}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Mixed type vectors are useful for combining semantically similar classes. Some examples of semantically related classes include time across different granularities (e.g. daily, monthly, annual) and probability distributions (e.g. Normal, Uniform, Poisson). These groups of vector types typically share common statistical operations which vary in results with the attributes of each vector. The 'vecvec' data structure facilitates efficient storage and computation across multiple vectors within the same object.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://pkg.mitchelloharawild.com/vecvec/}
  \item \url{https://github.com/mitchelloharawild/vecvec}
  \item Report bugs at \url{https://github.com/mitchelloharawild/vecvec/issues}
}

}
\author{
\strong{Maintainer}: Mitchell O'Hara-Wild \email{mail@mitchelloharawild.com} (\href{https://orcid.org/0000-0001-6729-7695}{ORCID})

Authors:
\itemize{
  \item Mitchell O'Hara-Wild \email{mail@mitchelloharawild.com} (\href{https://orcid.org/0000-0001-6729-7695}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hook.R
\name{vecvec_register}
\alias{vecvec_register}
\title{Register methods for a vecvec subclass}
\usage{
vecvec_register(x)
}
\arguments{
\item{x}{An S7 class object that extends \link{class_vecvec}. Typically the
class object created by \code{\link[S7:new_class]{S7::new_class()}} in your package, e.g.
\code{class_myvec}.}
}
\value{
\code{NULL}, invisibly. Called for its side effects.
}
\description{
Call \code{vecvec_register()} inside your package's \code{.onLoad()} function to
register the S3 methods that allow a custom
\link[=class_vecvec]{vecvec} subclass to participate in the
\href{https://vctrs.r-lib.org/}{vctrs} vector system.
}
\details{
Specifically, \code{vecvec_register()} registers:
\itemize{
\item \verb{vec_cast.<class>.*} methods, so that other types can be cast \emph{to} your
vecvec subclass.
\item \verb{vec_cast.*.<class>} methods, so that your vecvec subclass can be cast
\emph{from} other types.
}
}
\examples{
\dontrun{
# In your package, define a vecvec subclass:
class_myvec <- S7::new_class(
  "myvec",
  parent = vecvec::class_vecvec
)

# Then register it in .onLoad():
.onLoad <- function(libname, pkgname) {
  S7::methods_register()
  vecvec::vecvec_register(class_myvec)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vecvec.R
\name{is_vecvec}
\alias{is_vecvec}
\title{Test if an object is a vecvec}
\usage{
is_vecvec(x)
}
\arguments{
\item{x}{Object to test.}
}
\value{
\code{TRUE} if \code{x} inherits from \code{class_vecvec}, \code{FALSE} otherwise.
}
\description{
Test if an object is a vecvec
}
\examples{
vv <- vecvec(1:3, letters)
is_vecvec(vv)
is_vecvec(1:3)

}
\seealso{
\code{\link[=vecvec]{vecvec()}} to create a vecvec object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00_S7_classes.R
\name{class_vecvec}
\alias{class_vecvec}
\title{S7 class for vecvec}
\usage{
class_vecvec(x = list(), i = seq_len(sum(lengths(x))))
}
\arguments{
\item{x}{An unnamed list of arbitrary vectors. The vectors can be of
different types and lengths, but they must all be vectors (according to
\link[vctrs:vec_is]{vctrs::vec_is}). Adjacent vectors of the same type will be combined
together into a single vector.}

\item{i}{A vector of integers specifying the location of each element in \code{x}
as if they were combined in order. The values in \code{i} must be between 1 and
the total number of elements across all vectors in \code{x}, and can contain
duplicates. If not provided, it defaults to a sequence from 1 to the total
number of elements across all vectors in \code{x}.}
}
\value{
A vector of vectors of S7 object of class \code{class_vecvec}.
}
\description{
\code{class_vecvec()} constructs a new vecvec object from a list of vectors. It is
designed to be performant with minimal checks on the inputs. Direct use of
\code{class_vecvec()} is useful for developers, but for users it is recommended to
use \code{\link[=vecvec]{vecvec()}} to create vecvec objects.
}
\examples{
# Create a vecvec prototype
class_vecvec()

# Construct a vecvec from a list of vectors
class_vecvec(list(letters, rnorm(10)))

# Fully specify a vecvec with locations
class_vecvec(
  x = list(letters, rnorm(10)),
  i = c(1:3, 27:31, 26:4, 32:36)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vecvec.R
\name{vecvec}
\alias{vecvec}
\title{Create a vector of vectors}
\usage{
vecvec(...)
}
\arguments{
\item{...}{Vectors to combine. Each vector is stored as a separate typed
slot; no type coercion is performed.}
}
\value{
A \code{vecvec} object whose length equals the total number of elements
across all input vectors.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

A \code{vecvec} is a vector that holds elements of different types without
coercing them to a common type. Unlike a list of vectors, a \code{vecvec} behaves
as a flat vector (hence vector of vectors). This means that you can
operations (such as indexing, arithmetic, and statistics) apply across all
elements of a \code{vecvec} as if they were combined into a single vector.

Mixed-type vectors can be useful if you need to store heterogeneous data in a
vector-like structure. In most cases, I believe this is bad practice for data
analytics, but this could be useful for tidying up messy data. The most
valuable use case for vecvec is as a data structure for mixed-type semantic
vectors. This package is used by \href{https://pkg.mitchelloharawild.com/mixtime/}{mixtime}
and \href{https://pkg.mitchelloharawild.com/distributional/}{distributional} to
create vectors of time with different chronons and distributions with
different shapes.

To convert a \code{vecvec} back to a plain typed vector, use \code{\link[=unvecvec]{unvecvec()}}, which
casts all elements to a common type via \code{\link[vctrs:vec_cast]{vctrs::vec_cast()}}.
}
\examples{
# Mixed types are preserved without coercion
vv <- vecvec(Sys.Date(), rnorm(3), letters)
vv

# .[i] Indexing works like a flat vector
vv[c(1L, 3L, 7L)]

# .[[i]] drops to the original vector type
vv[[2L]]

}
\seealso{
\code{\link[=unvecvec]{unvecvec()}} to coerce a \code{vecvec} to a single-typed vector;
\code{\link[=is_vecvec]{is_vecvec()}} to test whether an object is a \code{vecvec}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apply.R
\name{vecvec_apply_fn}
\alias{vecvec_apply_fn}
\title{Function factory for vecvec_apply}
\usage{
vecvec_apply_fn(.f, ptype = NULL, SIMPLIFY = !is.null(ptype))
}
\arguments{
\item{.f}{A function to apply to each vector}

\item{ptype}{A prototype to simplify to. If \code{NULL}, the result will be a
\code{vecvec} object. If not \code{NULL}, the result will be simplified to the type
of \code{ptype} if possible.}

\item{SIMPLIFY}{If \code{TRUE}, the \verb{[unvecvec()]} will be applied to the result
of \code{vecvec_apply()}, using \code{ptype} as the target type. If \code{FALSE}, the
result will always be a \code{vecvec} object.}
}
\value{
A function that applies \code{.f} to each vector of a \code{vecvec} object and
optionally simplifies the result.
}
\description{
The \code{vecvec_apply_fn()} function is a function factory that creates applies
the function \code{.f} to each vector and optionally simplifies the result with
\verb{[unvecvec()]}. The function matches the forms of the original function
\code{.f} and can be used to define methods for generic functions that apply to
\code{vecvec} objects. If \code{.f} is a primitive function, the resulting function will
have a apply over an argument \code{x} and pass through \code{...}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apply.R
\name{vecvec_apply}
\alias{vecvec_apply}
\title{Apply a function to each vector of the vecvec}
\usage{
vecvec_apply(x, .f, ...)
}
\arguments{
\item{x}{A vecvec object}

\item{.f}{A function to apply to each vector}

\item{...}{Additional arguments passed to \code{.f}}
}
\value{
A vecvec data type with the same structure as \code{x} but with each
vector transformed by \code{.f}.
}
\description{
The \code{vecvec_apply()} function applies a function \code{.f} to each vector in the
\code{vecvec} vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statistic.R
\name{masked-cor}
\alias{masked-cor}
\alias{var}
\alias{cov}
\alias{cor}
\title{Correlation, Variance and Covariance}
\usage{
var(x, ...)

cov(x, ...)

cor(x, ...)
}
\arguments{
\item{x}{a numeric vector, matrix or data frame.}

\item{...}{Additional arguments passed to \code{\link[stats:var]{stats::var}}, \code{\link[stats:cov]{stats::cov}}, and
\code{\link[stats:cor]{stats::cor}}.}
}
\description{
\code{var}, \code{cov}, and \code{cor}  compute the variance of x and the covariance or
correlation of x and y if these are vectors. If x and y are matrices then
the covariances (or correlations) between the columns of x and the columns
of y are computed.

These functions mask and wrap the \code{\link[stats:var]{stats::var}}, \code{\link[stats:cov]{stats::cov}}, and
\code{\link[stats:cor]{stats::cor}} functions to add support for \code{vecvec} objects. More details
can be found in the documentation for those functions.
}
\seealso{
\code{\link[stats:var]{stats::var()}}, \code{\link[stats:cov]{stats::cov()}}, \code{\link[stats:cor]{stats::cor()}}
}
